#!/usr/bin/python3
#
# The Qubes OS Project, http://www.qubes-os.org
#
# Copyright (C) 2010  Rafal Wojtczuk  <rafal@invisiblethingslab.com>
#               2020  Norbert Kamiński  <norbert.kaminski@3mdeb.com>
#
# SPDX-License-Identifier: LGPL-2.1+
#

import base64
import glob
import hashlib
import itertools
import os
import shutil
import struct
import subprocess
import tempfile
from qubes_fwupd_common import EXIT_CODES, create_dirs

FWUPD_DOM0_DIR = "/var/cache/fwupd/qubes"
FWUPD_DOM0_UPDATES_DIR = os.path.join(FWUPD_DOM0_DIR, "updates")
FWUPD_DOM0_UNTRUSTED_DIR = os.path.join(FWUPD_DOM0_UPDATES_DIR, "untrusted")
FWUPD_DOM0_METADATA_DIR = os.path.join(FWUPD_DOM0_DIR, "metadata")

FWUPD_VM_DIR = "/home/user/.cache/fwupd"
FWUPD_VM_UPDATES_DIR = os.path.join(FWUPD_VM_DIR, "updates")
FWUPD_VM_METADATA_DIR = os.path.join(FWUPD_VM_DIR, "metadata")
FWUPD_PKI = "/etc/pki/fwupd"
FWUPD_PKI_PGP = "/etc/pki/fwupd/GPG-KEY-Linux-Vendor-Firmware-Service"
FWUPD_DOWNLOAD_PREFIX = "https://fwupd.org/downloads/"
HEADS_UPDATES_DIR = "/boot/updates"


class FwupdReceiveUpdates:
    def _check_shasum(self, file_path, sha):
        """Compares computed SHA256 checksum with `sha` parameter.

        Keyword arguments:
        file_path -- absolute path to the file
        sha -- SHA256 checksum of the file
        """
        with open(file_path, "rb") as f:
            c_sha = hashlib.sha256(f.read()).hexdigest()
        if c_sha != sha:
            self.clean_cache()
            raise ValueError(f"Computed checksum {c_sha} did NOT match {sha}.")

    def _jcat_verification(self, file_path, file_directory):
        """Verifies sha1 and sha256 checksum, GPG signature,
        and PKCS#7 signature.

        Keyword argument:
        file_path -- absolute path to jcat file
        file_directory -- absolute path to the directory to jcat file location
        """
        assert file_path.startswith("/"), f"bad file path {file_path!r}"
        cmd_jcat = ["jcat-tool", "verify", file_path, "--public-keys", FWUPD_PKI]
        p = subprocess.Popen(
            cmd_jcat, cwd=file_directory, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        stdout, __ = p.communicate()
        verification = stdout.decode("utf-8")
        print(verification)
        if p.returncode != 0:
            self.clean_cache()
            raise Exception("jcat-tool: Verification failed")

    def _crc24(self, data):
        """Calculate CRC-24

        Checksum calculation for PGP armored signature. This algorithm isn't
        available in Python standard library, but it's simple enough to
        implement here. It doesn't need to be fast, nor side-channel resistant.
        """
        crc = 0xB704CE  # crc24_init
        for b in data:
            crc ^= b << 16
            for i in range(8):
                crc <<= 1
                if crc & 0x1000000:
                    crc ^= 0x1864CFB  # crc24_poly
        return crc

    def _pgp_parse(self, signature_path):
        """Verifies if GPG signature is correctly formatted

        Verify if signature is well formed - sqv will verify if the signature
        itself correct, but may accept extra data in the signature packets
        that could later confuse GnuPG used by fwupd.
        Verify also armor checksum, as sqv doesn't do that.
        """

        # sigparse expects binary format, so decode base64 first
        with open(signature_path, "rb") as sig:
            data = sig.read(4096)  # arbitrary size limit
            if sig.read(1) != b"":
                raise Exception("pgp: signature too big")
            lines = data.splitlines()
            # format described in RFC-4880 ch 6
            if lines[0:2] != [b"-----BEGIN PGP SIGNATURE-----", b""]:
                raise Exception("pgp: invalid header")
            if lines[-1] != b"-----END PGP SIGNATURE-----":
                raise Exception("pgp: invalid footer")
            checksum = lines[-2]
            if checksum[0] != ord("=") or len(checksum) != 5:
                raise Exception("pgp: invalid checksum format")
            base64_data = b"".join(lines[2:-2])
            data = base64.b64decode(base64_data, validate=True)
            crc = base64.b64decode(checksum[1:], validate=True)
            crc = struct.unpack(">I", b"\0" + crc)[0]
            if crc != self._crc24(data):
                raise Exception("pgp: invalid checksum")
            with tempfile.NamedTemporaryFile() as tmp:
                tmp.write(data)
                tmp.flush()
                p = subprocess.Popen(["sigparse", tmp.name], stderr=subprocess.PIPE)
                _, stderr = p.communicate()
                if p.returncode != 0:
                    raise Exception("pgp: invalid signature format: " + stderr.decode())
        # reconstruct armored data to ensure its canonical form
        with open(signature_path, "wb") as sig:
            sig.write(b"-----BEGIN PGP SIGNATURE-----\n")
            sig.write(b"\n")
            encoded = iter(base64.b64encode(data))
            while line := bytes(itertools.islice(encoded, 76)):
                sig.write(line + b"\n")
            sig.write(checksum + b"\n")
            sig.write(b"-----END PGP SIGNATURE-----\n")

    def _pgp_verification(self, signature_path, file_path):
        """Verifies GPG signature.

        Keyword argument:
        signature_path -- absolute path to signature file
        file_path -- absolute path to the signed file location
        """
        assert file_path.startswith("/"), f"bad file path {file_path!r}"
        try:
            self._pgp_parse(signature_path)
        except Exception:
            self.clean_cache()
            raise
        cmd_verify = [
            "sqv",
            "--keyring",
            FWUPD_PKI_PGP,
            "--",
            signature_path,
            file_path,
        ]
        p = subprocess.Popen(cmd_verify, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, __ = p.communicate()
        signature_key = stdout.decode("utf-8")
        if p.returncode != 0 or not signature_key:
            self.clean_cache()
            raise Exception("pgp: Verification failed")

    def _reconstruct_jcat(self, jcat_path, file_path):
        """Reconstruct jcat file from verified parts

        Currently included parts: .asc, .sha256
        Hashes are generated locally if missing

        Arguments:
        jcat_path - absolute path to the output jcat file
        file_path - absolute path to the signed file
        """
        # generate missing hashes
        hashes = ("sha256",)
        for hash_ext in hashes:
            hash_fname = f"{file_path}.{hash_ext}"
            if not os.path.exists(hash_fname):
                # TODO: switch to hashlib.file_digest (py3.11)
                with open(file_path, "rb") as f_data:
                    hash_val = hashlib.new(hash_ext, f_data.read()).hexdigest()
                with open(hash_fname, "w") as f_hash:
                    f_hash.write(hash_val)

        signatures = ("asc",)

        file_id = os.path.basename(file_path)
        for sign_type in signatures + hashes:
            sign_path = f"{file_path}.{sign_type}"
            if not os.path.exists(sign_path):
                raise Exception(f"Missing signature: {sign_path}")
            jcat_cmd = ["jcat-tool", "import", jcat_path, file_id, sign_path]
            subprocess.check_call(jcat_cmd)

    def handle_fw_update(self, updatevm, sha, filename):
        """Copies firmware update archives from the updateVM.

        Keyword arguments:
        updatevm -- update VM name
        sha -- SHA256 checksum of the firmware update archive
        filename -- name of the firmware update archive
        """
        create_dirs(FWUPD_DOM0_UPDATES_DIR, FWUPD_DOM0_UNTRUSTED_DIR)

        with tempfile.TemporaryDirectory(dir=FWUPD_DOM0_UNTRUSTED_DIR) as tmpdir:
            dom0_firmware_untrusted_path = os.path.join(tmpdir, filename)
            updatevm_firmware_file_path = os.path.join(FWUPD_VM_UPDATES_DIR, filename)

            cmd_copy = [
                "qvm-run",
                "--pass-io",
                "--no-gui",
                "-q",
                "-a",
                "--no-shell",
                "--",
                updatevm,
                "cat",
                "--",
                updatevm_firmware_file_path,
            ]
            with open(dom0_firmware_untrusted_path, "bx") as untrusted_file:
                p = subprocess.Popen(cmd_copy, stdout=untrusted_file, shell=False)
            p.wait()
            if p.returncode != 0:
                raise Exception("qvm-run: Copying firmware file failed!!")

            self._check_shasum(dom0_firmware_untrusted_path, sha)
            # jcat verification will be done by fwupd itself
            self.arch_name = filename
            self.arch_path = os.path.join(FWUPD_DOM0_UPDATES_DIR, filename)
            shutil.move(dom0_firmware_untrusted_path, self.arch_path)

    def handle_metadata_update(self, updatevm, metadata_url):
        """Copies metadata files from the updateVM.

        Keyword argument:
        updatevm -- update VM name
        """
        metadata_name = os.path.basename(metadata_url)
        self.metadata_file = os.path.join(FWUPD_DOM0_METADATA_DIR, metadata_name)
        self.metadata_file_jcat = self.metadata_file + ".jcat"
        self.metadata_file_updatevm = os.path.join(FWUPD_VM_METADATA_DIR, metadata_name)
        create_dirs(FWUPD_DOM0_METADATA_DIR, FWUPD_DOM0_UNTRUSTED_DIR)
        with tempfile.TemporaryDirectory(dir=FWUPD_DOM0_UNTRUSTED_DIR) as tmpdir:
            cmd_copy_metadata_file = [
                "qvm-run",
                "--pass-io",
                "--no-gui",
                "--no-shell",
                "--",
                updatevm,
                "cat",
                "--",
                self.metadata_file_updatevm,
            ]
            # TODO: switch to ed25519 once firmware.xml.gz.jcat will have it
            cmd_copy_metadata_file_signature = [
                "qvm-run",
                "--pass-io",
                "--no-gui",
                "--no-shell",
                "--",
                updatevm,
                "cat",
                "--",
                self.metadata_file_updatevm + ".asc",
            ]
            untrusted_metadata_file = os.path.join(tmpdir, metadata_name)

            with open(untrusted_metadata_file, "bx") as untrusted_file_1, open(
                untrusted_metadata_file + ".asc", "bx"
            ) as untrusted_file_2, subprocess.Popen(
                cmd_copy_metadata_file, stdout=untrusted_file_1
            ) as p, subprocess.Popen(
                cmd_copy_metadata_file_signature, stdout=untrusted_file_2
            ) as q:
                p.wait()
                q.wait()
            if p.returncode != 0:
                raise Exception("qvm-run: Copying metadata file failed!!")
            if q.returncode != 0:
                raise Exception("qvm-run: Copying metadata signature failed!!")

            self._pgp_verification(
                untrusted_metadata_file + ".asc", untrusted_metadata_file
            )
            self._reconstruct_jcat(
                untrusted_metadata_file + ".jcat", untrusted_metadata_file
            )
            # verified, move into trusted dir
            shutil.move(untrusted_metadata_file, self.metadata_file)
            shutil.move(untrusted_metadata_file + ".jcat", self.metadata_file_jcat)

    def clean_cache(self):
        """Removes updates data"""
        print("Cleaning dom0 cache directories")
        if os.path.exists(FWUPD_DOM0_METADATA_DIR):
            shutil.rmtree(FWUPD_DOM0_METADATA_DIR)
        if os.path.exists(FWUPD_DOM0_UPDATES_DIR):
            shutil.rmtree(FWUPD_DOM0_UPDATES_DIR)
        if os.path.exists(HEADS_UPDATES_DIR):
            shutil.rmtree(HEADS_UPDATES_DIR)



#!/usr/bin/python3
# pylint: disable=invalid-name,missing-docstring
#
# Copyright (C) 2023 Collabora Ltd
# Author: Frédéric Danis <frederic.danis@collabora.com>
#
# SPDX-License-Identifier: LGPL-2.1+

import argparse
import base64
import json
import os
import subprocess
import sys
from typing import Any, Dict, List, Optional, Tuple
from zipfile import ZipFile, ZIP_DEFLATED

URB_INTERRUPT = 1
URB_CONTROL = 2
URB_BULK = 3

DESCRIPTOR_DEVICE = "1"
DESCRIPTOR_CONFIGURATION = "2"
DESCRIPTOR_STRING = "3"
DESCRIPTOR_INTERFACE = "4"
DESCRIPTOR_ENDPOINT = "5"
DESCRIPTOR_EXTRA = "33"

INTERFACE_CLASS_HID = 3
INTERFACE_CLASS_SMARTCARD = 11

CCID_PC_TO_RDR_SET_PARAMETERS = 0x61
CCID_PC_TO_RDR_ICC_POWER_ON = 0x62
CCID_PC_TO_RDR_ICC_POWER_OFF = 0x63
CCID_PC_TO_RDR_GET_SLOT_STATUS = 0x65
CCID_PC_TO_RDR_ESCAPE = 0x6B
CCID_PC_TO_RDR_TRANSFER_BLOCK = 0x6F
CCID_RDR_TO_PC_DATA_BLOCK = 0x80
CCID_RDR_TO_PC_SLOT_STATUS = 0x81
CCID_RDR_TO_PC_PARAMETERS = 0x82
CCID_RDR_TO_PC_ESCAPE = 0x83


def get_int(data: str) -> int:
    if data[:2] == "0x":
        return int(data, 16)
    return int(data)


def add_bytes(array: bytearray, string: str, size: int) -> None:
    array += get_int(string).to_bytes(length=size, byteorder="little")


class Pcap2Emulation:
    def __init__(self, device_ids: str):
        self.device: Dict[str, Any] = {}
        self.platform_id = ""
        self.phases: List[Any] = []
        self.device_ids: List[List[str]] = []
        self.interface_index = 0
        self.endpoint_index = 0
        self.previous_data: Optional[str]
        self.bulk_incoming_lens: Dict[str, int] = {}
        self.usb_port = None
        self.enumerate = False

        for i in range(len(device_ids)):
            device_id = device_ids[i].split(":")
            if len(device_id) > 2:
                sys.stderr.write(f"Malformed device ID: {device_ids[i]}\n\n")
                exit(1)
            if len(device_id) == 2 and len(device_id[1]) == 0:
                del device_id[1]
            if device_id not in self.device_ids:
                self.device_ids.append(device_id)

    def _save_phase(self) -> None:
        self.phases.append({"UsbDevices": [self.device]})
        self.interface_index = 0
        self.endpoint_index = 0

    def save_archive(self, path: str) -> None:
        if not self.phases:
            return

        print(f"Found {len(self.phases)} phases:")
        phase = 0

        if path.endswith(".zip"):
            emulation_file = path
        else:
            emulation_file = path + ".zip"
        with ZipFile(emulation_file, "w", compression=ZIP_DEFLATED) as write_file:
            print(f"- phase {phase} as setup.json")
            json_string = json.dumps(
                self.phases[phase], indent=2, separators=(",", " : ")
            )
            write_file.writestr("setup.json", json_string)
            phase += 1

            if len(self.phases) > 2:
                print(f"- phase {phase} as install.json")
                json_string = json.dumps(
                    self.phases[phase], indent=2, separators=(",", " : ")
                )
                write_file.writestr("install.json", json_string)
                phase += 1

            print(f"- phase {phase} as reload.json")
            json_string = json.dumps(
                self.phases[phase], indent=2, separators=(",", " : ")
            )
            write_file.writestr("reload.json", json_string)
            phase += 1

        print("Emulation file saved to " + emulation_file)

        while phase < len(self.phases):
            phase_path = f"{path}-{phase}.json"
            with open(phase_path, "w") as dump_file:
                json.dump(
                    self.phases[phase],
                    dump_file,
                    indent=2,
                    separators=(",", " : "),
                )
                print(f"- unused phase {phase} saved to {phase_path}")
            phase += 1

    def _run_tshark(self, file: str, tshark_filter: str) -> Any:
        cmd = ["tshark", "-n", "-T", "ek", "-l", "-2", "-r", file, "-R"]
        print("running: " + " ".join(cmd) + ' "' + tshark_filter + '"')
        cmd.append(tshark_filter)
        return subprocess.Popen(cmd, stdout=subprocess.PIPE)

    def _get_usb_addrs(self, file: str) -> Tuple[str, List[str]]:
        tshark_filter = ""
        for i in range(len(self.device_ids)):
            if len(tshark_filter) == 0:
                tshark_filter += "("
            else:
                tshark_filter += " or ("
            tshark_filter += "usb.idVendor == 0x" + self.device_ids[i][0]
            if len(self.device_ids[i]) == 2 and len(self.device_ids[i][1]) > 0:
                tshark_filter += " and usb.idProduct == 0x" + self.device_ids[i][1]
            tshark_filter += ")"

        usb_bus = ""
        usb_addrs: List[str] = []
        p = self._run_tshark(file, tshark_filter)
        for line in p.stdout:
            pcap_data = json.loads(line)
            if "layers" in pcap_data:
                if not usb_bus:
                    usb_bus = pcap_data["layers"]["usb"]["usb_usb_bus_id"]
                elif usb_bus != pcap_data["layers"]["usb"]["usb_usb_bus_id"]:
                    print(
                        "* Warning: Found different USB Bus ID: expected {}, found {}".format(
                            usb_bus, pcap_data["layers"]["usb"]["usb_usb_bus_id"]
                        )
                    )
                addr = pcap_data["layers"]["usb"]["usb_usb_device_address"]
                if addr not in usb_addrs:
                    usb_addrs.append(addr)
        return usb_bus, usb_addrs

    def _get_interrupt_event(self, layers: Dict[str, Any]) -> Dict[str, str]:
        if "usb_usb_capdata" in layers:
            captured_data = str(
                base64.b64encode(
                    bytes.fromhex(layers["usb_usb_capdata"].replace(":", ""))
                ),
                "utf-8",
            )
            s = "InterruptTransfer:Endpoint=0x{:02x}".format(
                get_int(layers["usb"]["usb_usb_endpoint_address"])
            )
            if layers["usb"]["usb_usb_endpoint_address_direction"] == "1":
                if hasattr(self, "previous_data") and self.previous_data:
                    s += f",Data={self.previous_data}"
                    self.previous_data = None
                else:
                    s += ",Data="
            else:
                self.previous_data = captured_data
                s += f",Data={captured_data}"
            s += f",Length=0x{int(layers['usb']['usb_usb_data_len']):x}"
            return {"Id": s, "Data": captured_data}
        return {}

    def _get_bulk_event(self, layers: Dict[str, Any]) -> Dict[str, str]:
        captured_data = None
        if "usbccid" in layers:
            message_type = get_int(layers["usbccid"]["usbccid_usbccid_bMessageType"])
            ccid = bytearray([message_type])
            add_bytes(ccid, layers["usbccid"]["usbccid_usbccid_dwLength"], 4)
            add_bytes(ccid, layers["usbccid"]["usbccid_usbccid_bSlot"], 1)
            add_bytes(ccid, layers["usbccid"]["usbccid_usbccid_bSeq"], 1)

            if message_type == CCID_PC_TO_RDR_SET_PARAMETERS:
                add_bytes(ccid, layers["usbccid"]["usbccid_usbccid_bProtocolNum"], 1)
                add_bytes(
                    ccid, layers["usbccid"]["usbccid_usbccid_hf_ccid_Reserved"], 2
                )
                ccid += bytearray.fromhex(
                    layers["data"]["data_data_data"].replace(":", "")
                )
            elif message_type == CCID_PC_TO_RDR_ICC_POWER_ON:
                add_bytes(ccid, layers["usbccid"]["usbccid_usbccid_bPowerSelect"], 1)
                add_bytes(
                    ccid, layers["usbccid"]["usbccid_usbccid_hf_ccid_Reserved"], 2
                )
            elif (
                message_type == CCID_PC_TO_RDR_ICC_POWER_OFF
                or message_type == CCID_PC_TO_RDR_GET_SLOT_STATUS
            ):
                add_bytes(
                    ccid, layers["usbccid"]["usbccid_usbccid_hf_ccid_Reserved"], 3
                )
            elif message_type == CCID_PC_TO_RDR_ESCAPE:
                ccid += bytearray.fromhex(
                    layers["usbccid"]["usbccid_usbccid_abRFU"].replace(":", "")
                )
                ccid += bytearray.fromhex(
                    layers["data"]["data_data_data"].replace(":", "")
                )
            elif message_type == CCID_PC_TO_RDR_TRANSFER_BLOCK:
                add_bytes(ccid, layers["usbccid"]["usbccid_usbccid_bBWI"], 1)
                add_bytes(ccid, layers["usbccid"]["usbccid_usbccid_wLevelParameter"], 2)
                ccid += bytearray.fromhex(
                    layers["data"]["data_data_data"].replace(":", "")
                )
            elif message_type == CCID_RDR_TO_PC_DATA_BLOCK:
                add_bytes(ccid, layers["usbccid"]["usbccid_usbccid_bStatus"], 1)
                add_bytes(ccid, layers["usbccid"]["usbccid_usbccid_bError"], 1)
                add_bytes(ccid, layers["usbccid"]["usbccid_usbccid_bChainParameter"], 1)
                if "data" in layers:
                    ccid += bytearray.fromhex(
                        layers["data"]["data_data_data"].replace(":", "")
                    )
            elif message_type == CCID_RDR_TO_PC_SLOT_STATUS:
                add_bytes(ccid, layers["usbccid"]["usbccid_usbccid_bStatus"], 1)
                add_bytes(ccid, layers["usbccid"]["usbccid_usbccid_bError"], 1)
                add_bytes(ccid, layers["usbccid"]["usbccid_usbccid_bClockStatus"], 1)
            elif message_type == CCID_RDR_TO_PC_PARAMETERS:
                add_bytes(ccid, layers["usbccid"]["usbccid_usbccid_bStatus"], 1)
                add_bytes(ccid, layers["usbccid"]["usbccid_usbccid_bError"], 1)
                add_bytes(ccid, layers["usbccid"]["usbccid_usbccid_bProtocolNum"], 1)
            elif message_type == CCID_RDR_TO_PC_ESCAPE:
                add_bytes(ccid, layers["usbccid"]["usbccid_usbccid_bStatus"], 1)
                add_bytes(ccid, layers["usbccid"]["usbccid_usbccid_bError"], 1)
                add_bytes(ccid, layers["usbccid"]["usbccid_usbccid_bRFU"], 1)
                ccid += bytearray.fromhex(
                    layers["data"]["data_data_data"].replace(":", "")
                )
            else:
                print(f"Unknown USB CCID Bulk message type: 0x{message_type:02X}")

            captured_data = str(base64.b64encode(ccid), "utf-8")
        elif "usb_usb_capdata" in layers:
            captured_data = str(
                base64.b64encode(
                    bytes.fromhex(layers["usb_usb_capdata"].replace(":", ""))
                ),
                "utf-8",
            )

        if captured_data:
            s = "BulkTransfer:Endpoint=0x{:02x}".format(
                get_int(layers["usb"]["usb_usb_endpoint_address"])
            )
            if layers["usb"]["usb_usb_endpoint_address_direction"] == "1":
                if layers["usb"]["usb_usb_request_in"] in self.bulk_incoming_lens:
                    length = self.bulk_incoming_lens[
                        layers["usb"]["usb_usb_request_in"]
                    ]
                else:
                    length = int(layers["usb"]["usb_usb_data_len"])
                data = str(
                    base64.b64encode(bytes.fromhex("00" * length)),
                    "utf-8",
                )
                s += f",Data={data}"
            else:
                length = int(layers["usb"]["usb_usb_data_len"])
                s += f",Data={captured_data}"
            s += f",Length=0x{length:x}"
            return {"Id": s, "Data": captured_data}
        elif layers["usb"]["usb_usb_endpoint_address_direction"] == "1":
            if "usb_usb_urb_len" in layers["usb"]:
                self.bulk_incoming_lens[layers["frame"]["frame_frame_number"]] = int(
                    layers["usb"]["usb_usb_urb_len"]
                )
        return {}

    def _get_interface_descriptor(
        self, layers: Dict[str, Any], descriptor_index: int
    ) -> Any:
        table = {
            "usb_usb_bInterfaceNumber": "InterfaceNumber",
            "usb_usb_bAlternateSetting": "AlternateSetting",
            "usb_usb_bNumEndpoints": "NumEndpoints",
            "usb_usb_bInterfaceClass": "InterfaceClass",
            "usb_usb_bInterfaceSubClass": "InterfaceSubClass",
            "usb_usb_bInterfaceProtocol": "InterfaceProtocol",
            "usb_usb_iInterface": "Interface",
        }

        # Interface descriptor frame may occur multiple times,
        # it should not change unless device has been re-enumerated
        if len(layers["usb_usb_bInterfaceNumber"]) <= self.interface_index:
            return None

        interface: Dict[str, Any] = {
            "Length": get_int(layers["usb_usb_bLength"][descriptor_index]),
            "DescriptorType": 4,
        }
        for key in table:
            # data can be a string or a list of strings
            if type(layers[key]) is str:
                val = get_int(layers[key])
            else:
                val = get_int(layers[key][self.interface_index])

            if key not in layers:
                continue

            interface[table[key]] = val
            if key == "usb_usb_bNumEndpoints":
                interface["UsbEndpoints"] = []

        return interface

    def _get_endpoint_descriptor(
        self, layers: Dict[str, Any], index: int
    ) -> Dict[str, int]:
        table = {
            "usb_usb_bEndpointAddress": "EndpointAddress",
            "usb_usb_bInterval": "Interval",
            "usb_usb_wMaxPacketSize": "MaxPacketSize",
        }

        endpoint = {
            "DescriptorType": 5,
        }
        for key in table:
            val = get_int(layers[key][index])
            if val != 0:
                endpoint[table[key]] = val
        return endpoint

    def _get_hid_descriptor(self, layers: Dict[str, Any], index: int) -> str:
        desc = bytearray()
        add_bytes(desc, layers["usb_usb_bLength"][index], 1)
        add_bytes(desc, layers["usb_usb_bDescriptorType"][index], 1)
        add_bytes(desc, layers["usbhid_usbhid_descriptor_hid_bcdHID"], 2)
        add_bytes(desc, layers["usbhid_usbhid_descriptor_hid_bCountryCode"], 1)
        add_bytes(desc, layers["usbhid_usbhid_descriptor_hid_bNumDescriptors"], 1)
        add_bytes(desc, layers["usbhid_usbhid_descriptor_hid_bDescriptorType"], 1)
        add_bytes(desc, layers["usbhid_usbhid_descriptor_hid_wDescriptorLength"], 2)
        return str(base64.b64encode(desc), "utf-8")

    def _get_smartcard_descriptor(self, layers: Dict[str, Any], index: int) -> str:
        desc = bytearray()
        add_bytes(desc, layers["usb_usb_bLength"][index], 1)
        add_bytes(desc, layers["usb_usb_bDescriptorType"][index], 1)
        add_bytes(desc, layers["usbccid_usbccid_bcdCCID"], 2)
        add_bytes(desc, layers["usbccid_usbccid_bMaxSlotIndex"], 1)
        add_bytes(desc, layers["usbccid_usbccid_bVoltageSupport"], 1)
        add_bytes(desc, layers["usbccid_usbccid_dwProtocols"], 4)
        add_bytes(desc, layers["usbccid_usbccid_dwDefaultClock"], 4)
        add_bytes(desc, layers["usbccid_usbccid_dwMaximumClock"], 4)
        add_bytes(desc, layers["usbccid_usbccid_bNumClockSupported"], 1)
        add_bytes(desc, layers["usbccid_usbccid_dwDataRate"], 4)
        add_bytes(desc, layers["usbccid_usbccid_dwMaxDataRate"], 4)
        add_bytes(desc, layers["usbccid_usbccid_bNumDataRatesSupported"], 1)
        add_bytes(desc, layers["usbccid_usbccid_dwMaxIFSD"], 4)
        add_bytes(desc, layers["usbccid_usbccid_dwSynchProtocols"], 4)
        add_bytes(desc, layers["usbccid_usbccid_dwMechanical"], 4)
        add_bytes(desc, layers["usbccid_usbccid_dwFeatures"], 4)
        add_bytes(desc, layers["usbccid_usbccid_dwMaxCCIDMessageLength"], 4)
        add_bytes(desc, layers["usbccid_usbccid_hf_ccid_bClassGetResponse"], 1)
        add_bytes(desc, layers["usbccid_usbccid_hf_ccid_bClassEnvelope"], 1)
        add_bytes(desc, layers["usbccid_usbccid_hf_ccid_wLcdLayout"], 2)
        add_bytes(desc, layers["usbccid_usbccid_hf_ccid_bPINSupport"], 1)
        add_bytes(desc, layers["usbccid_usbccid_hf_ccid_bMaxCCIDBusySlots"], 1)
        return str(base64.b64encode(desc), "utf-8")

    def _save_event(self, event: Dict[str, str]) -> None:
        if not self.device:
            return

        self.device["UsbEvents"].append(event)

    def parse_file(self, file: str) -> None:
        bus_id, addrs = self._get_usb_addrs(file)
        if len(addrs) == 0:
            print("Device(s) not found in pcap file")
            return

        # Filter the device related packets and the C_PORT_CONNECTION clear
        # feature packets to allow detection of the re-plug/re-enumerate events
        tshark_filter = f"usb.bus_id == {bus_id}"
        tshark_filter += " and (usbhub.setup.PortFeatureSelector == 16"
        for addr in addrs:
            tshark_filter += f" or usb.device_address == {addr}"
        tshark_filter += ")"

        p = self._run_tshark(file, tshark_filter)
        for line in p.stdout:
            pcap_data = json.loads(line)
            if "layers" in pcap_data:
                layers = pcap_data["layers"]
                usb_port = None

                if (
                    layers["frame"]["frame_frame_cap_len"]
                    != layers["frame"]["frame_frame_len"]
                ):
                    print(
                        "* Incomplete frame {}: {} bytes captured < {}".format(
                            layers["frame"]["frame_frame_number"],
                            layers["frame"]["frame_frame_cap_len"],
                            layers["frame"]["frame_frame_len"],
                        )
                    )

                # Store the USB port of C_PORT_CONNECTION clear feature packet
                # Trigger an enumeration requirement if it's the same port that
                # was used for the previous device description
                if "usbhub_usbhub_setup_Port" in layers:
                    usb_port = layers["usbhub_usbhub_setup_Port"]
                    if usb_port == self.usb_port:
                        self.enumerate = True

                if get_int(layers["usb"]["usb_usb_transfer_type"]) == URB_INTERRUPT:
                    event = self._get_interrupt_event(layers)
                    if len(event) > 0:
                        self._save_event(event)

                elif get_int(layers["usb"]["usb_usb_transfer_type"]) == URB_CONTROL:
                    if "usb_usb_bDescriptorType" in layers:
                        descriptor_index = -1

                        # usb_usb_bDescriptorType can be a string or a list of strings
                        if type(layers["usb_usb_bDescriptorType"]) is str:
                            layer = [layers["usb_usb_bDescriptorType"]]
                        else:
                            layer = layers["usb_usb_bDescriptorType"]

                        for descriptor_type in layer:
                            descriptor_index += 1

                            if descriptor_type == DESCRIPTOR_DEVICE:
                                # Check this is a reply for the Vid[:Pid] expected
                                if layers["usb"]["usb_usb_src"] == "host":
                                    continue
                                found = False
                                for i in range(len(self.device_ids)):
                                    if get_int(layers["usb_usb_idVendor"]) == int(
                                        self.device_ids[i][0], 16
                                    ):
                                        if len(self.device_ids[i]) == 1 or get_int(
                                            layers["usb_usb_idProduct"]
                                        ) == int(self.device_ids[i][1], 16):
                                            found = True
                                            break
                                if not found:
                                    continue

                                # Save the previous USB device
                                if (
                                    self.device
                                    and len(self.device["UsbInterfaces"]) > 0
                                ):
                                    self._save_phase()

                                # Create a new USB device
                                # using a fake PlatformId based on USB bus id and device address,
                                # this PlatformId should be stable for all recorded devices
                                if not self.platform_id:
                                    self.platform_id = "usb:{:02x}:{:02x}".format(
                                        int(layers["usb"]["usb_usb_bus_id"]),
                                        int(layers["usb"]["usb_usb_device_address"]),
                                    )
                                # Device re-enumeration is triggered when 'Created' time differs
                                # from previous phase, this keeps the 'Created' time from previous
                                # phase unless a re-enumeration requirement has been detected
                                if not self.device or self.enumerate:
                                    frame_time = layers["frame"]["frame_frame_time"]
                                    self.usb_port = usb_port
                                    self.enumerate = False
                                else:
                                    frame_time = self.device["Created"]
                                self.device = {
                                    "PlatformId": self.platform_id,
                                    "Created": frame_time,
                                    "Tags": ["org.freedesktop.fwupd.emulation.v1"],
                                    "IdVendor": int(layers["usb_usb_idVendor"]),
                                    "IdProduct": int(layers["usb_usb_idProduct"], 16),
                                    "Device": int(layers["usb_usb_bcdDevice"], 16),
                                    "USB": int(layers["usb_usb_bcdUSB"], 16),
                                    "Manufacturer": int(
                                        layers["usb_usb_iManufacturer"]
                                    ),
                                    "DeviceClass": int(layers["usb_usb_bDeviceClass"]),
                                    "DeviceSubClass": int(
                                        layers["usb_usb_bDeviceSubClass"]
                                    ),
                                    "DeviceProtocol": int(
                                        layers["usb_usb_bDeviceProtocol"]
                                    ),
                                    "Product": int(layers["usb_usb_iProduct"]),
                                    "SerialNumber": int(
                                        layers["usb_usb_iSerialNumber"]
                                    ),
                                    "UsbInterfaces": [],
                                    "UsbEvents": [],
                                }

                            if descriptor_type == DESCRIPTOR_CONFIGURATION:
                                if "usb_usb_iConfiguration" in layers:
                                    # The GetConfigurationIndex GUsb event is not directly
                                    # related to a specific USB event, but data can be
                                    # retrieved from the DESCRIPTOR CONFIGURATION request
                                    index = (
                                        layers["usb_usb_iConfiguration"]
                                        .encode("utf-8")
                                        .hex()
                                    )
                                    event = {
                                        "Id": "GetConfigurationIndex",
                                        "Data": str(
                                            base64.b64encode(bytes.fromhex(index)),
                                            "utf-8",
                                        ),
                                    }
                                    self._save_event(event)

                            if descriptor_type == DESCRIPTOR_STRING:
                                if "usb_usb_DescriptorIndex" in layers:
                                    desc_index = get_int(
                                        layers["usb_usb_DescriptorIndex"]
                                    )
                                    if desc_index == 0:
                                        # The list of supported languages are not recorded for the emulation
                                        continue
                                    event_str = {
                                        "Id": "GetStringDescriptor:DescIndex=0x{:02x}".format(
                                            desc_index
                                        )
                                    }
                                    # duplicate the event so it can also be used for GetStringDescriptorBytes
                                    language_id = get_int(layers["usb_usb_LanguageId"])
                                    length = get_int(layers["usb_usb_setup_wLength"])
                                    event_bytes = {
                                        "Id": "GetStringDescriptorBytes:DescIndex=0x{:02x}".format(
                                            desc_index
                                        )
                                    }
                                    event_bytes["Id"] += f",Langid=0x{language_id:04x}"
                                    event_bytes["Id"] += f",Length=0x{length:x}"

                                elif "usb_usb_bString" in layers:
                                    if int(layers["usb_usb_bLength"]) != len(
                                        layers["usb_usb_bString"].encode("utf-16")
                                    ):
                                        # Discard frame used to retrieve STRING DESCRIPTOR length
                                        continue

                                    # Found a new STRING DESCRIPTOR response
                                    data = (
                                        layers["usb_usb_bString"].encode("utf-8").hex()
                                    )
                                    if "usb_usb_capdata" in layers:
                                        # Add leftover capture data
                                        data += "00"
                                        data += layers["usb_usb_capdata"].replace(
                                            ":", ""
                                        )
                                    event_str["Data"] = str(
                                        base64.b64encode(bytes.fromhex(data)), "utf-8"
                                    )

                                    # now that the event is completed it can be added to the device events
                                    self._save_event(event_str)
                                    # duplicate the event so it can also be used for GetStringDescriptorBytes
                                    event_bytes["Data"] = event_str["Data"]
                                    self._save_event(event_bytes)

                            if descriptor_type == DESCRIPTOR_INTERFACE:
                                interface = self._get_interface_descriptor(
                                    layers, descriptor_index
                                )
                                if not interface:
                                    continue

                                # Add the interface to the device
                                self.device["UsbInterfaces"].append(interface)

                                if (
                                    "InterfaceClass" in interface
                                    and "InterfaceSubClass" in interface
                                    and "InterfaceProtocol" in interface
                                ):
                                    # The GetCustomIndex GUsb event is not directly
                                    # related to a specific USB event, but data can be
                                    # retrieved from the DESCRIPTOR INTERFACE request
                                    index = interface["Interface"].to_bytes(1, "big")
                                    event = {
                                        "Id": "GetCustomIndex:ClassId=0x{:02x},SubclassId=0x{:02x},ProtocolId=0x{:02x}".format(
                                            interface["InterfaceClass"],
                                            interface["InterfaceSubClass"],
                                            interface["InterfaceProtocol"],
                                        ),
                                        "Data": str(
                                            base64.b64encode(index),
                                            "utf-8",
                                        ),
                                    }
                                    self._save_event(event)

                                self.interface_index += 1

                            if descriptor_type == DESCRIPTOR_ENDPOINT:
                                if (
                                    len(layers["usb_usb_bEndpointAddress"])
                                    > self.endpoint_index
                                ):
                                    endpoint = self._get_endpoint_descriptor(
                                        layers, self.endpoint_index
                                    )

                                    # Add the endpoint to the first interface with missing endpoint
                                    for interface in self.device["UsbInterfaces"]:
                                        if (
                                            "UsbEndpoints" in interface
                                            and len(interface["UsbEndpoints"])
                                            < interface["NumEndpoints"]
                                        ):
                                            interface["UsbEndpoints"].append(endpoint)
                                            break
                                    self.endpoint_index += 1

                            if descriptor_type == DESCRIPTOR_EXTRA:
                                for interface in self.device["UsbInterfaces"]:
                                    if (
                                        interface["InterfaceClass"]
                                        == INTERFACE_CLASS_HID
                                    ):
                                        d = self._get_hid_descriptor(
                                            layers, descriptor_index
                                        )
                                        interface["ExtraData"] = d
                                        break
                                    elif (
                                        interface["InterfaceClass"]
                                        == INTERFACE_CLASS_SMARTCARD
                                    ):
                                        d = self._get_smartcard_descriptor(
                                            layers, descriptor_index
                                        )
                                        interface["ExtraData"] = d
                                        break

                    elif (
                        "usb_usb_bmRequestType_type" in layers
                        and int(layers["usb_usb_bmRequestType_type"], 16) == URB_CONTROL
                    ):
                        # Found vendor CONTROL URB request
                        direction = not (layers["usb_usb_bmRequestType_direction"])
                        s = f"ControlTransfer:Direction=0x{direction:02x}"
                        s += ",RequestType=0x{:02x}".format(
                            int(layers["usb_usb_bmRequestType_type"], 16)
                        )
                        s += ",Recipient=0x{:02x}".format(
                            int(layers["usb_usb_bmRequestType_recipient"], 16)
                        )
                        s += f",Request=0x{int(layers['usb_usb_setup_bRequest']):02x}"
                        s += f",Value=0x{int(layers['usb_usb_setup_wValue'], 16):04x}"
                        s += f",Idx=0x{int(layers['usb_usb_setup_wIndex']):04x}"
                        if "usb_usb_data_fragment" in layers:
                            data = layers["usb_usb_data_fragment"].replace(":", "")
                        else:
                            data = "00" * int(layers["usb_usb_setup_wLength"])
                        s += ",Data={}".format(
                            str(base64.b64encode(bytes.fromhex(data)), "utf-8")
                        )
                        s += f",Length=0x{int(layers['usb_usb_setup_wLength']):x}"
                        event = {"Id": s}

                        if direction:
                            # Duplicate the outgoing data as response and add the event to the device events
                            event["Data"] = str(
                                base64.b64encode(bytes.fromhex(data)), "utf-8"
                            )
                            self._save_event(event)

                    elif "usb_usb_control_Response" in layers:
                        # Found CONTROL URB response
                        data = layers["usb_usb_control_Response"].replace(":", "")
                        event["Data"] = str(
                            base64.b64encode(bytes.fromhex(data)), "utf-8"
                        )

                        # now that the event is complete it can be added to the device events
                        self._save_event(event)

                elif get_int(layers["usb"]["usb_usb_transfer_type"]) == URB_BULK:
                    # TODO: check it
                    event = self._get_bulk_event(layers)
                    if event:
                        self._save_event(event)

                else:
                    if not self.device:
                        continue

                    print(
                        "Unknown frame type: " + layers["usb"]["usb_usb_transfer_type"]
                    )

        # Save the last USB device
        if "UsbInterfaces" in self.device:
            self._save_phase()


if __name__ == "__main__":
    options = argparse.ArgumentParser(description="Convert pcap file to emulation file")
    options.add_argument("input_pcap", type=str, help="pcap file to convert")
    options.add_argument("output_archive", type=str, help="Output archive path")
    options.add_argument(
        "device_id",
        metavar=("VendorID[:ProductID]"),
        type=str,
        nargs="+",
        help="Device ID in hexadecimal",
    )

    args = options.parse_args()

    path = os.path.abspath(os.path.expanduser(os.path.expandvars(args.input_pcap)))
    parser = Pcap2Emulation(args.device_id)
    parser.parse_file(path)
    parser.save_archive(args.output_archive)



#!/usr/bin/python3
# pylint: disable=invalid-name,missing-docstring
#
# Copyright 2024 Richard Hughes <richard@hughsie.com>
#
# SPDX-License-Identifier: LGPL-2.1-or-later

import struct
import sys
import binascii

from pcapng import FileScanner


def parse_pkt(block, title: str, offset: int):
    (
        st_id,
        st_type,
        st_size,
        rc_ctrl,
        rc_sts,
        rc_offset,
        rc_length,
    ) = struct.unpack_from("<BBBxxBBLL", block.packet_data, offset=offset)
    buf: bytes = block.packet_data[offset + 15 : offset + 15 + rc_length]
    if title == "GET" and rc_ctrl & 0x80:
        title = "get"
    print(
        f"{title} {st_id}:{st_type}:{st_size} "
        f"CTRL:0x{rc_ctrl:02X} STS:0x{rc_sts:02X} "
        f"OFFSET:0x{rc_offset:08X} LENGTH:0x{rc_length:02X}, "
        f"DATA:{binascii.hexlify(buf)}"
    )


def parse_file(filename: str):
    with open(filename, "rb") as f:
        scanner = FileScanner(f)
        for block in scanner:
            try:
                # print(block)
                if block.packet_len == 98:
                    parse_pkt(block, title="SET", offset=36)
                elif block.packet_len == 90:
                    parse_pkt(block, title="GET", offset=28)
            except AttributeError:
                pass


if __name__ == "__main__":
    parse_file(sys.argv[1])



#!/usr/bin/python3
#
# Copyright (C) 2021 Richard Hughes <richard@hughsie.com>
#
# SPDX-License-Identifier: LGPL-2.1+

import glob
import sys


def test_files() -> int:

    rc: int = 0

    for fn in glob.glob("**/*.quirk", recursive=True):
        with open(fn, "r") as f:
            for line in f.read().split("\n"):
                if line.startswith(" ") or line.endswith(" "):
                    print(f"{fn} has leading or trailing whitespace: {line}")
                    rc = 1
                    continue
                if not line or line.startswith("#"):
                    continue
                if line.startswith("["):
                    if not line.endswith("]"):
                        print(f"{fn} has invalid section header: {line}")
                        rc = 1
                        continue
                    for deprecated in ["DeviceInstanceId", "Guid"]:
                        if line.find(deprecated) != -1:
                            print(f"{fn} has deprecated prefix: {deprecated}")
                            rc = 1
                            continue
                else:
                    sections = line.split(" = ")
                    if len(sections) != 2:
                        print(f"{fn} has invalid line: {line}")
                        rc = 1
                        continue
                    for section in sections:
                        if section.strip() != section:
                            print(f"{fn} has invalid spacing: {line}")
                            rc = 1
                            break
    return rc


if __name__ == "__main__":

    # all done!
    sys.exit(test_files())



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017 Peter Jones <pjones@redhat.com>
# Copyright (C) 2020 Richard Hughes <richard@hughsie.com>
#
# SPDX-License-Identifier: LGPL-2.1+
#
# pylint: disable=wrong-import-position,too-many-locals,unused-argument,too-many-statements
# pylint: disable=invalid-name,too-many-instance-attributes,missing-module-docstring
# pylint: disable=missing-function-docstring,missing-class-docstring,too-few-public-methods

import os
import sys
import argparse
import tarfile
import math
import io
import struct

from typing import Dict, Optional, Any

import cairo
import gi

gi.require_version("Pango", "1.0")
gi.require_version("PangoCairo", "1.0")
from gi.repository import Pango, PangoCairo


def languages(podir: str):
    for x in open(os.path.join(podir, "LINGUAS"), "r").readlines():
        yield x.strip()
    yield "en"


class PotFile:
    def __init__(self, fn=None):
        self.msgs: Dict[str, str] = {}
        if fn:
            self.parse(fn)

    def parse(self, fn: str) -> None:
        with open(fn, "r") as f:
            lang_en: Optional[str] = None
            for line in f.read().split("\n"):
                if not line:
                    continue
                if line[0] == "#":
                    continue
                try:
                    key, value = line.split(" ", maxsplit=1)
                except ValueError:
                    continue
                if key == "msgid":
                    lang_en = value[1:-1]
                    continue
                if key == "msgstr" and lang_en:
                    self.msgs[lang_en] = value[1:-1]
                    lang_en = None
                    continue


def _cairo_surface_write_to_bmp(img: cairo.ImageSurface) -> bytes:

    data = bytes(img.get_data())
    return (
        b"BM"
        + struct.pack(
            "<ihhiiiihhiiiiii",
            54 + len(data),  # size of BMP file
            0,  # unused
            0,  # unused
            54,  # pixel array offset
            40,  # DIB header
            img.get_width(),  # width
            img.get_height(),  # height
            1,  # planes
            32,  # BPP
            0,  # no compression
            len(data),  # size of the raw bitmap data
            2835,  # 72DPI H
            2835,  # 72DPI V
            0,  # palette
            0,  # all colors are important
        )
        + data
    )


def main(args) -> int:

    # open output archive
    with tarfile.open(args.out, "w:xz", preset=8) as tar:

        for lang in languages(args.podir):
            # these are the 1.6:1 of some common(ish) screen widths
            if lang == "en":
                label_translated: str = args.label
            else:
                potfile = PotFile(os.path.join(args.podir, f"{lang}.po"))
                try:
                    label_translated = potfile.msgs[args.label]
                except KeyError:
                    continue
                if label_translated == args.label:
                    continue
            for width, height in (
                (640, 480),
                (800, 600),
                (1024, 768),
                (1280, 720),
                (1280, 800),
                (1366, 768),
                (1536, 864),
                (1600, 900),
                (1920, 1080),
                (1920, 1200),
                (2160, 1350),
                (2560, 1440),
                (3840, 2160),
                (5120, 2880),
                (5688, 3200),
                (7680, 4320),
            ):

                # generate PangoLanguage
                font_desc = f"Sans {height / 32:.2f}px"
                fd = Pango.FontDescription(font_desc)
                font_option = cairo.FontOptions()
                font_option.set_antialias(cairo.ANTIALIAS_SUBPIXEL)
                l = Pango.Language.from_string(lang)

                # create surface
                img = cairo.ImageSurface(cairo.FORMAT_RGB24, 1, 1)
                cctx = cairo.Context(img)
                layout = PangoCairo.create_layout(cctx)
                pctx = layout.get_context()
                pctx.set_font_description(fd)
                pctx.set_language(l)
                fs = pctx.load_fontset(fd, l)
                PangoCairo.context_set_font_options(pctx, font_option)

                attrs = Pango.AttrList()
                length = len(bytes(label_translated, "utf8"))
                items = Pango.itemize(pctx, label_translated, 0, length, attrs, None)
                if not items:
                    continue
                try:
                    # urgh, https://gitlab.gnome.org/GNOME/pango/-/merge_requests/829
                    # -- if we depend on Pango >= 1.56.2 we can drop the fallback
                    #
                    # ...or if we depend on python3-gobject >= 3.51.0 we can use the nicer:
                    #    len(inspect.signature(Pango.shape).parameters)
                    gs = Pango.shape(label_translated, length, items[0].analysis)
                except TypeError:
                    gs = Pango.GlyphString()
                    Pango.shape(label_translated, length, items[0].analysis, gs)
                del img, cctx, pctx, layout

                def find_size(fs, f, data):
                    """find our size, I hope..."""
                    (ink, log) = gs.extents(f)
                    if ink.height == 0 or ink.width == 0:
                        return False
                    data.update({"log": log, "ink": ink})
                    return True

                data: Dict[str, Any] = {}
                fs.foreach(find_size, data)
                if len(data) == 0:
                    print("Missing sans fonts")
                    return 2
                log = data["log"]
                ink = data["ink"]

                surface_height = math.ceil(max(ink.height, log.height) / Pango.SCALE)
                surface_width = math.ceil(max(ink.width, log.width) / Pango.SCALE)

                x = -math.ceil(log.x / Pango.SCALE)
                y = -math.ceil(log.y / Pango.SCALE)

                img = cairo.ImageSurface(
                    cairo.FORMAT_RGB24, surface_width, surface_height
                )
                cctx = cairo.Context(img)
                layout = PangoCairo.create_layout(cctx)
                pctx = layout.get_context()
                PangoCairo.context_set_font_options(pctx, font_option)

                cctx.set_source_rgb(1, 1, 1)
                cctx.move_to(x, y - surface_height / 2)

                def do_write(fs, f, data):
                    """write out glyphs"""
                    ink = gs.extents(f)[0]
                    if ink.height == 0 or ink.width == 0:
                        return False
                    PangoCairo.show_glyph_string(cctx, f, gs)
                    return True

                # flip the image to write the bitmap upside-down
                mat = cairo.Matrix()
                mat.scale(1, -1)
                cctx.transform(mat)

                fs.foreach(do_write, None)
                img.flush()

                # convert to BMP and add to archive
                with io.BytesIO() as io_bmp:
                    io_bmp.write(_cairo_surface_write_to_bmp(img))
                    filename = f"fwupd-{lang}-{width}-{height}.bmp"
                    tarinfo = tarfile.TarInfo(filename)
                    tarinfo.size = io_bmp.tell()
                    io_bmp.seek(0)
                    tar.addfile(tarinfo, fileobj=io_bmp)

    # success
    return 0


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Make UX images")
    parser.add_argument("--label", help="Update text", required=True)
    parser.add_argument("--podir", help="Po location", required=True)
    parser.add_argument("--out", help="Output archive", required=True)
    sys.exit(main(parser.parse_args()))



#!/usr/bin/python3
# pylint: disable=invalid-name,missing-docstring
#
# Copyright (C) 2022 Richard Hughes <richard@hughsie.com>
#
# SPDX-License-Identifier: LGPL-2.1+

import argparse
import datetime
import sys
import json
import os

from typing import Dict, List, Any

import xml.etree.ElementTree as ET

if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--version", type=str, help="fwupd version")
    parser.add_argument("-s", "--schema-version", type=str, help="schema version")
    parser.add_argument(
        "filename_dst", action="store", type=str, help="XML destination"
    )
    parser.add_argument("json_attrs", nargs="+", help="JSON attributes")
    args = parser.parse_args()

    # parse JSON
    items: List[str] = []
    for fn in sorted(args.json_attrs):
        try:
            with open(fn, "rb") as f:
                item = json.loads(f.read())
        except json.decoder.JSONDecodeError as e:
            print(f"failed to parse {fn}: {str(e)}")
            sys.exit(1)
        for tag in ["id", "name", "failure-results"]:
            if tag not in item:
                print("skipping {} as no {}".format(fn), tag)
                continue
        items.append(item)

    oval_definitions = ET.Element("oval_definitions")
    oval_definitions.set("xmlns", "http://oval.mitre.org/XMLSchema/oval-definitions-5")
    oval_definitions.set("xmlns:oval", "http://oval.mitre.org/XMLSchema/oval-common-5")
    oval_definitions.set(
        "xmlns:unix-def", "http://oval.mitre.org/XMLSchema/oval-definitions-5#unix"
    )
    oval_definitions.set(
        "xmlns:red-def", "http://oval.mitre.org/XMLSchema/oval-definitions-5#linux"
    )
    oval_definitions.set(
        "xmlns:ind-def",
        "http://oval.mitre.org/XMLSchema/oval-definitions-5#independent",
    )
    oval_definitions.set("xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance")
    oval_definitions.set(
        "xsi:schemaLocation",
        "http://oval.mitre.org/XMLSchema/oval-common-5 oval-common-schema.xsd "
        "http://oval.mitre.org/XMLSchema/oval-definitions-5 oval-definitions-schema.xsd "
        "http://oval.mitre.org/XMLSchema/oval-definitions-5#unix unix-definitions-schema.xsd "
        "http://oval.mitre.org/XMLSchema/oval-definitions-5#linux linux-definitions-schema.xsd",
    )

    generator = ET.SubElement(oval_definitions, "generator")
    for key, value in {
        "oval:product_name": "fwupd",
        "oval:product_version": args.version,
        "oval:schema_version": args.schema_version,
        "oval:timestamp": datetime.datetime.now().isoformat(),
    }.items():
        oval = ET.SubElement(generator, key)
        oval.text = value

    definitions = ET.SubElement(oval_definitions, "definitions")
    for item in items:

        definition = ET.SubElement(definitions, "definition")
        definition.set("class", "patch")
        definition.set("id", f"oval:{item['id']}:def:1")
        definition.set("version", "1")

        metadata = ET.SubElement(definition, "metadata")

        ET.SubElement(metadata, "title").text = item["name"]

        affected = ET.SubElement(metadata, "affected")
        affected.set("family", "unix")
        ET.SubElement(affected, "platform").text = "All"  # is this valid?

        if "issues" in item:
            for issue in item["issues"]:
                reference = ET.SubElement(metadata, "reference")
                reference.set("ref_id", issue)
                if issue.startswith("CVE-"):
                    reference.set(
                        "ref_url", f"https://nvd.nist.gov/vuln/detail/{issue}"
                    )
                    reference.set("source", "CVE")

        if "description" in item:
            ET.SubElement(metadata, "description").text = "\n".join(item["description"])

        criteria = ET.SubElement(definition, "criteria")
        criteria.set("operator", "OR")
        criterion = ET.SubElement(criteria, "criterion")
        criterion.set("comment", item["name"])
        criterion.set("test_ref", f"oval:{item['id']}:tst:1")

    tests = ET.SubElement(oval_definitions, "tests")
    for item in items:

        red_def = ET.SubElement(tests, "red-def:fwupdsecattr_test")
        red_def.set("check", "at least one")
        red_def.set("comment", item["name"])
        red_def.set("id", f"oval:{item['id']}:tst:1")
        red_def.set("version", "1")

        red_def_object = ET.SubElement(red_def, "red-def:object")
        red_def_object.set("object_ref", f"oval:{item['id']}:obj:1")
        red_def_state = ET.SubElement(red_def, "red-def:state")
        red_def_state.set("state_ref", f"oval:{item['id']}:ste:1")

    objects = ET.SubElement(oval_definitions, "objects")
    for item in items:

        red_def = ET.SubElement(objects, "red-def:fwupdsecattr_object")
        red_def.set("id", f"oval:{item['id']}:obj:1")
        red_def.set("version", "1")

        red_def_stream_id = ET.SubElement(red_def, "red-def:stream-id")
        red_def_stream_id.set("datatype", "string")
        red_def_stream_id.text = format(item["id"])

    states = ET.SubElement(oval_definitions, "states")
    for item in items:

        red_def = ET.SubElement(states, "red-def:fwupdsecattr_state")
        red_def.set("id", f"oval:{item['id']}:ste:1")
        red_def.set("version", "1")

        red_def_security_attr = ET.SubElement(red_def, "red-def:security-attr")
        red_def_security_attr.set("datatype", "string")
        red_def_security_attr.set("operation", "pattern match")
        red_def_security_attr.text = "|".join(
            [value for value, _ in item["failure-results"].items()]
        )

    ET.indent(oval_definitions, space=" ", level=0)
    with open(args.filename_dst, "wb") as f:
        f.write(ET.tostring(oval_definitions, encoding="utf-8", xml_declaration=True))



#!/usr/bin/python3
#
# Copyright (C) 2017 Richard Hughes <richard@hughsie.com>
#
# SPDX-License-Identifier: LGPL-2.1+
#
# pylint: disable=invalid-name,missing-docstring,consider-using-f-string

import json
import sys

from typing import Dict, List, Any

import gi
from gi.repository import GLib

gi.require_version("Fwupd", "2.0")
gi.require_version("Json", "1.0")

from gi.repository import Fwupd  # pylint: disable=wrong-import-position
from gi.repository import Json  # pylint: disable=wrong-import-position


def _minimize_json(json_str: str) -> str:
    nodes = json.loads(json_str)
    new_attrs: List[Dict[str, Any]] = []
    new_devices: List[Dict[str, Any]] = []
    new_bios_settings: List[Dict[str, Any]] = []
    try:
        for attr in nodes["SecurityAttributes"]:
            new_attr: Dict[str, Any] = {}
            for key in attr:
                if key in ["AppstreamId", "HsiResult", "Flags", "Plugin"]:
                    new_attr[key] = attr[key]
            new_attrs.append(new_attr)
    except KeyError:
        pass
    try:
        for device in nodes["Devices"]:
            new_device: Dict[str, Any] = {}
            for key in device:
                if key not in ["Created", "Modified", "Releases", "Plugin"]:
                    new_device[key] = device[key]
            new_devices.append(new_device)
    except KeyError:
        pass
    try:
        for device in nodes["BiosSettings"]:
            new_attr: Dict[str, Any] = {}
            for key in device:
                if key not in ["Filename"]:
                    new_attr[key] = device[key]
            new_bios_settings.append(new_attr)
    except KeyError:
        pass
    return json.dumps(
        {
            "SecurityAttributes": new_attrs,
            "Devices": new_devices,
            "BiosSettings": new_bios_settings,
        },
        indent=2,
        separators=(",", " : "),
    )


def _get_host_devices_and_attrs() -> str:

    # connect to the running daemon
    client = Fwupd.Client()
    builder = Json.Builder()
    builder.begin_object()

    # add devices
    try:
        devices = client.get_devices()
    except GLib.GError as e:
        print(f"ignoring {e}")
    else:
        builder.set_member_name("Devices")
        builder.begin_array()
        for device in devices:
            builder.begin_object()
            device.to_json_full(builder, Fwupd.DEVICE_FLAG_TRUSTED)
            builder.end_object()
        builder.end_array()

    # add security attributes
    try:
        attrs = client.get_host_security_attrs()
    except GLib.GError as e:
        print(f"ignoring {e}")
    else:
        builder.set_member_name("SecurityAttributes")
        builder.begin_array()
        for attr in attrs:
            builder.begin_object()
            attr.to_json(builder)
            builder.end_object()
        builder.end_array()

    # add BIOS settings
    try:
        attrs = client.get_bios_settings()
    except GLib.GError as e:
        print(f"ignoring {e}")
    else:
        builder.set_member_name("BiosSettings")
        builder.begin_array()
        for attr in attrs:
            builder.begin_object()
            attr.to_json(builder)
            builder.end_object()
        builder.end_array()

    # export to JSON
    builder.end_object()
    generator = Json.Generator()
    generator.set_pretty(True)
    generator.set_root(builder.get_root())
    return generator.to_data()[0]


if len(sys.argv) < 2:
    sys.stdout.write(_minimize_json(sys.stdin.read()))
else:
    for fn in sys.argv[1:]:

        try:
            with open(fn, "rb") as f_in:
                json_in = f_in.read().decode()
        except FileNotFoundError:
            json_in = _get_host_devices_and_attrs()
        json_out = _minimize_json(json_in).encode()
        with open(fn, "wb") as f_out:
            f_out.write(json_out)



#!/usr/bin/python3
# pylint: disable=invalid-name,missing-docstring
#
# Copyright (C) 2022 Richard Hughes <richard@hughsie.com>
#
# SPDX-License-Identifier: LGPL-2.1+

import argparse
import sys
import json
import os
from typing import Dict, List, Any

if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "filename_src", action="store", type=str, help="markdown source"
    )
    parser.add_argument(
        "filename_dst", action="store", type=str, help="markdown destination"
    )
    parser.add_argument("json_attrs", nargs="+", help="JSON attributes")
    args = parser.parse_args()

    with open(args.filename_src, "rb") as f:
        template = f.read()

    txt: List[str] = []
    for fn in sorted(args.json_attrs):
        try:
            with open(fn, "rb") as f:
                item = json.loads(f.read())
        except json.decoder.JSONDecodeError as e:
            print(f"failed to parse {fn}: {str(e)}")
            sys.exit(1)
        if "id" not in item:
            print(f"skipping {fn} as no id")
            continue
        txt += [f"<a id=\"{item['id']}\"></a>"]
        if "deprecated-ids" in item:
            for deprecated_id in item["deprecated-ids"]:
                txt += [f'<a id="{deprecated_id}"></a>']
        if "name" in item:
            txt += [f"### [{item['name']}](#{item['id']})"]
        if "description" in item:
            for para in item["description"]:
                txt += [para]
        if "failure-impact" in item:
            txt += ["**Impact:**"]
            for para in item["failure-impact"]:
                txt += [para]
        if "failure-results" in item and "success-results" in item:
            txt += ["**Possible results:**"]
            tmp: List[str] = []
            for value, desc in item["failure-results"].items():
                tmp += [f"- `{value}`: {desc} (failure)"]
            for value, desc in item["success-results"].items():
                tmp += [f"- `{value}`: {desc} (success)"]
            txt += ["\n".join(tmp)]
        if "hsi-level" in item and "fwupd-version" in item:
            txt += [
                "A test success result is needed to meet HSI-{} on "
                "systems that run this test. *[v{}]*".format(
                    item["hsi-level"], item["fwupd-version"]
                )
            ]
        if "resolution" in item:
            txt += [f"**Resolution:** {item['resolution']}"]
        if "issues" in item:
            txt += ["**Issues:**"]
            tmp: List[str] = []
            for issue in item["issues"]:
                if issue.startswith("CVE-"):
                    tmp += [f"- [{issue}](https://nvd.nist.gov/vuln/detail/{issue})"]
                else:
                    tmp += [f"- {issue}"]
            txt += ["\n".join(tmp)]
        if "references" in item:
            txt += ["**References:**"]
            tmp: List[str] = []
            for url, title in item["references"].items():
                tmp += [f"- [{title}]({url})"]
            txt += ["\n".join(tmp)]
        if "requires" in item:
            txt += ["**Hardware requirements:**"]
            if "CPUID\\VID_GenuineIntel" in item["requires"]:
                txt += ["This attribute will only be available when using Intel CPUs."]
            elif "CPUID\\VID_AuthenticAMD" in item["requires"]:
                txt += ["This attribute will only be available when using AMD CPUs."]
        if "more-information" in item:
            txt += ["**More information:**"]
            for para in item["more-information"]:
                txt += [para]

    with open(args.filename_dst, "wb") as f:
        f.write(template.decode().replace("{{tests}}", "\n\n".join(txt)).encode())



#!/usr/bin/python3
#
# The Qubes OS Project, http://www.qubes-os.org
#
# Copyright (C) 2021  Norbert Kamiński  <norbert.kaminski@3mdeb.com>
#
# SPDX-License-Identifier: LGPL-2.1+
#

import io
import os
import platform
import shutil
import imp
import src.qubes_fwupd_heads as qf_heads
import sys
import unittest

from test.fwupd_logs import HEADS_XML

CUSTOM_METADATA = "https://fwupd.org/downloads/firmware-3c81bfdc9db5c8a42c09d38091944bc1a05b27b0.xml.gz"
QUBES_FWUPDMGR_REPO = "./src/qubes_fwupdmgr.py"
QUBES_FWUPDMGR_BINDIR = "/usr/sbin/qubes-fwupdmgr"


class TestQubesFwupdHeads(unittest.TestCase):
    def setUp(self):
        if os.path.exists(QUBES_FWUPDMGR_REPO):
            self.qfwupd = imp.load_source("qubes_fwupdmgr", QUBES_FWUPDMGR_REPO)
        elif os.path.exists(QUBES_FWUPDMGR_BINDIR):
            self.qfwupd = imp.load_source("qubes_fwupdmgr", QUBES_FWUPDMGR_BINDIR)
        self.q = qf_heads.FwupdHeads()
        self.maxDiff = 2000
        self.captured_output = io.StringIO()
        sys.stdout = self.captured_output

    @unittest.skipUnless("qubes" in platform.release(), "Requires Qubes OS")
    def test_get_hwids(self):
        self.q._get_hwids()
        self.assertNotEqual(self.q.dom0_hwids_info, "")

    def test_gather_firmware_version_empty(self):
        self.q.dom0_hwids_info = ""
        return_code = self.q._gather_firmware_version()
        self.assertEqual(return_code, self.qfwupd.EXIT_CODES["NOTHING_TO_DO"])

    def test_gather_firmware_version(self):
        self.q.dom0_hwids_info = "CBET4000 Heads-v0.2.2-917-g19f0e65"
        self.q._gather_firmware_version()
        self.assertEqual(self.q.heads_version, "0.2.2-917-g19f0e65")

    @unittest.skipUnless("qubes" in platform.release(), "Requires Qubes OS")
    def test_parse_metadata(self):
        qmgr = self.qfwupd.QubesFwupdmgr()
        qmgr.metadata_file = CUSTOM_METADATA.replace(
            "https://fwupd.org/downloads", self.qfwupd.FWUPD_DOM0_METADATA_DIR
        )
        qmgr._download_metadata(metadata_url=CUSTOM_METADATA)
        self.q._parse_metadata(qmgr.metadata_file)
        self.assertTrue(self.q.metadata_info)

    def test_check_heads_updates_default_heads(self):
        self.q.metadata_info = HEADS_XML
        self.q.heads_version = "heads"
        return_code = self.q._parse_heads_updates("x230")
        self.assertEqual(return_code, self.qfwupd.EXIT_CODES["SUCCESS"])
        self.assertEqual(
            self.q.heads_update_url,
            "https://fwupd.org/downloads/e747a435bf24fd6081b77b6704b39cec5fa2dcf62e0ca6b86d8a6460121a1d07-heads_coreboot_x230-v0_2_3.cab",
        )
        self.assertEqual(
            self.q.heads_update_sha,
            "ba519a7a5d8136c8ade0cf0c775c58f3165f42798ff631c3f57f075897ef1586",
        )
        self.assertEqual(self.q.heads_update_version, "0.2.3")

    def test_check_heads_updates_no_updates(self):
        self.q.metadata_info = HEADS_XML
        self.q.heads_version = "0.2.3"
        return_code = self.q._parse_heads_updates("x230")
        self.assertEqual(return_code, self.qfwupd.EXIT_CODES["NOTHING_TO_DO"])

    def test_check_heads_updates_lower_version(self):
        self.q.metadata_info = HEADS_XML
        self.q.heads_version = "0.2.2"
        return_code = self.q._parse_heads_updates("x230")
        self.assertEqual(return_code, self.qfwupd.EXIT_CODES["SUCCESS"])
        self.assertEqual(
            self.q.heads_update_url,
            "https://fwupd.org/downloads/e747a435bf24fd6081b77b6704b39cec5fa2dcf62e0ca6b86d8a6460121a1d07-heads_coreboot_x230-v0_2_3.cab",
        )
        self.assertEqual(
            self.q.heads_update_sha,
            "ba519a7a5d8136c8ade0cf0c775c58f3165f42798ff631c3f57f075897ef1586",
        )
        self.assertEqual(self.q.heads_update_version, "0.2.3")

    @unittest.skipUnless("qubes" in platform.release(), "Requires Qubes OS")
    def test_copy_heads_firmware(self):
        qmgr = self.qfwupd.QubesFwupdmgr()
        self.q.heads_update_url = "https://fwupd.org/downloads/e747a435bf24fd6081b77b6704b39cec5fa2dcf62e0ca6b86d8a6460121a1d07-heads_coreboot_x230-v0_2_3.cab"
        self.q.heads_update_sha = (
            "ba519a7a5d8136c8ade0cf0c775c58f3165f42798ff631c3f57f075897ef1586"
        )
        self.q.heads_update_version = "0.2.3"
        qmgr._download_firmware_updates(
            self.q.heads_update_url, self.q.heads_update_sha
        )
        heads_boot_path = os.path.join(
            qf_heads.HEADS_UPDATES_DIR, self.q.heads_update_version
        )
        if os.path.exists(heads_boot_path):
            shutil.rmtree(heads_boot_path)
        ret_code = self.q._copy_heads_firmware(qmgr.arch_path)
        self.assertNotEqual(ret_code, self.qfwupd.EXIT_CODES["NOTHING_TO_DO"])
        firmware_path = os.path.join(heads_boot_path, "firmware.rom")
        self.assertTrue(os.path.exists(firmware_path))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/python3
#
# The Qubes OS Project, http://www.qubes-os.org
#
# Copyright (C) 2021  Norbert Kaminski  <norbert.kaminski@3mdeb.com>
#
# SPDX-License-Identifier: LGPL-2.1+
#

import json
import os
import re
import shutil
import subprocess
import tempfile
import sys
import xml.etree.ElementTree as ET

from pathlib import Path
from packaging import version as pversion

FWUPD_QUBES_DIR = "/usr/share/qubes-fwupd"

# Check if script is run by tests and append sys path properly
if __name__ == "__main__":
    sys.path.append(os.path.join(FWUPD_QUBES_DIR, "src"))
else:
    sys.path.append("./src")

try:
    from qubes_fwupd_heads import FwupdHeads
    from qubes_fwupd_update import FwupdUpdate, run_in_tty
    from fwupd_receive_updates import FwupdReceiveUpdates
    from qubes_fwupd_common import EXIT_CODES, create_dirs
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "qubes-fwupd modules not found.  You may need to reinstall package."
    )

FWUPD_DOM0_DIR = "/var/cache/fwupd/qubes"
FWUPD_DOM0_METADATA_DIR = os.path.join(FWUPD_DOM0_DIR, "metadata")
FWUPD_DOM0_UPDATES_DIR = os.path.join(FWUPD_DOM0_DIR, "updates")
FWUPD_DOWNLOAD_PREFIX = "https://fwupd.org/downloads/"
METADATA_URL = "https://fwupd.org/downloads/firmware.xml.xz"
METADATA_URL_JCAT = "https://fwupd.org/downloads/firmware.xml.xz.jcat"

FWUPDMGR = "/bin/fwupdmgr"

BIOS_UPDATE_FLAG = os.path.join(FWUPD_DOM0_DIR, "bios_update")
LVFS_TESTING_DOM0_FLAG = os.path.join(FWUPD_DOM0_DIR, "lvfs_testing")

HELP = {
    "Usage": [
        {
            "Command": "qubes-fwupdmgr [OPTION…][FLAG..]",
            "Example": "qubes-fwupdmgr refresh --whonix --url=<url>\n",
        }
    ],
    "Options": [
        {
            "get-devices": "Get all devices that support firmware updates",
            "get-updates": "Get the list of updates for connected hardware",
            "refresh": "Refresh metadata from remote server",
            "update": "Update chosen device to latest firmware version",
            "update-heads": "Updates heads firmware to the latest version (EXPERIMENTAL, not fully supported yet)",
            "downgrade": "Downgrade chosen device to chosen firmware version",
            "clean": "Delete all cached update files\n",
        }
    ],
    "Flags": [
        {
            "--whonix": "Download firmware updates via Tor",
            "--device": "Specify device for heads update (default - x230)",
            "--url": "Address of the custom metadata remote server\n",
        }
    ],
    "Help": [{"-h --help": "Show help options\n"}],
}


class QubesFwupdmgr(FwupdHeads, FwupdUpdate, FwupdReceiveUpdates):
    def _download_metadata(self, whonix=False, metadata_url=None):
        """Initialize downloading metadata files.

        Keywords arguments:
        whonix -- Flag enforces downloading the metadata updates via Tor
        metadata_url -- Download metadata from the custom url
        """
        if not metadata_url:
            raise Exception("missing metadata URL")
        self.download_metadata(whonix=whonix, metadata_url=metadata_url)
        self.handle_metadata_update(self.updatevm, metadata_url=metadata_url)
        if not os.path.exists(self.metadata_file):
            raise FileNotFoundError("Metadata file does not exist")

    def get_remotes(self):
        """Get metadata URLs for all enabled remotes"""

        if hasattr(self, "_remotes_cache"):
            return self._remotes_cache

        remotes_json = subprocess.check_output(
            [
                FWUPDMGR,
                "get-remotes",
                "--json",
            ]
        ).decode()
        remotes_list = json.loads(remotes_json)["Remotes"]
        remotes = {}
        for remote in remotes_list:
            name = remote["Id"]
            # skip disabled
            if remote.get("Enabled", "true") != "true":
                continue
            # skip local - for metadata refresh, we only care about those
            # actually needing refreshing
            if remote.get("Kind") != "download":
                continue
            # skip unsupported keyring kind
            if remote.get("KeyringKind") not in ("jcat",):
                print(
                    "Skipping remote '{}' due to unsupported keyring type '{}'".format(
                        name, remote.get("KeyringKind")
                    )
                )
            assert "MetadataUri" in remote
            remotes[name] = remote["MetadataUri"]

        self._remotes_cache = remotes
        return self._remotes_cache

    def refresh_metadata(self, whonix=False, metadata_url=None, remote_name=None):
        """Updates metadata with downloaded files.

        Keyword arguments:
        whonix -- Flag enforces downloading the metadata updates via Tor
        metadata_url -- Use custom metadata from the url
        remote_name -- Set refreshed metadata to this remote
        """
        if not metadata_url:
            if remote_name:
                metadata_url = self.get_remotes()[remote_name]
            else:
                raise Exception("missing metadata URL")
        metadata_name = os.path.basename(metadata_url)
        self.metadata_file = os.path.join(FWUPD_DOM0_METADATA_DIR, metadata_name)
        self.metadata_file_jcat = self.metadata_file + ".jcat"
        if not remote_name:
            if "testing" in metadata_url:
                remote_name = "lvfs-testing"
            else:
                remote_name = "lvfs"
        self._download_metadata(whonix=whonix, metadata_url=metadata_url)
        cmd_refresh = [
            FWUPDMGR,
            "refresh",
            self.metadata_file,
            self.metadata_file_jcat,
            remote_name,
        ]
        p = subprocess.Popen(cmd_refresh, stdout=subprocess.PIPE)
        output = p.communicate()[0].decode()
        print(output)
        if p.returncode != 0:
            raise Exception("fwupd-qubes: Refresh failed")
        if not output != "Successfully refreshed metadata manually":
            raise Exception("Manual metadata refresh failed!!!")

    def refresh_metadata_all(self, whonix=False):
        """Refresh metadata for all 'download' remotes

        Keyword arguments:
        whonix -- Flag enforces downloading the metadata updates via Tor
        """
        for name, url in self.get_remotes().items():
            try:
                self.refresh_metadata(whonix=whonix, remote_name=name, metadata_url=url)
            except Exception as e:
                print(f"Failed to refresh remote '{name}': {e}")

    def _get_dom0_updates(self):
        """Gathers information about available updates."""
        cmd_get_dom0_updates = [FWUPDMGR, "--json", "get-updates"]
        p = subprocess.Popen(cmd_get_dom0_updates, stdout=subprocess.PIPE)
        self.dom0_updates_info = p.communicate()[0].decode()
        if p.returncode != 0 and p.returncode != 2:
            raise Exception("fwupd-qubes: Getting available updates failed")

    def _parse_dom0_updates_info(self, updates_info):
        """Creates dictionary and list with information about updates.

        Keywords argument:
        updates_info - gathered update information
        """
        self.dom0_updates_info_dict = json.loads(updates_info)
        self.dom0_updates_list = [
            {
                "Name": device["Name"],
                "Version": device["Version"],
                "Releases": [
                    {
                        "Version": update["Version"],
                        "Url": update["Uri"],
                        "Checksum": update["Checksum"][-1],
                        "Description": update["Description"],
                    }
                    for update in device["Releases"]
                ],
            }
            for device in self.dom0_updates_info_dict["Devices"]
        ]

    def _download_firmware_updates(self, url, sha, whonix=False):
        """Initializes downloading firmware update archive.

        Keywords arguments:
        url -- url path to the firmware update archive
        sha -- SHA256 checksum of the firmware update archive
        whonix -- Flag enforces downloading the updates via Tor
        """
        self.cached = False
        self.download_firmware_updates(url, sha, whonix=whonix)
        if not self.cached:
            self.handle_fw_update(self.updatevm, sha, self.arch_name)
        if not os.path.exists(self.arch_path):
            raise FileNotFoundError("Firmware update files do not exist")

    def _user_input(self, updates_list, downgrade=False):
        """UI for update process.

        Keywords arguments:
        updates_dict - list of updates for specified device
        downgrade -- downgrade flag
        """
        decorator = "======================================================"
        if len(updates_list) == 0:
            print("No updates available.")
            return -EXIT_CODES["NOTHING_TO_DO"]
        if downgrade:
            print("Available downgrades:")
        else:
            print("Available updates:")
        self._updates_crawler(updates_list)

        while True:
            try:
                print("If you want to abandon process press 'N'.")
                choice = input("Otherwise choose a device number: ")
                if choice == "N" or choice == "n":
                    return -EXIT_CODES["NOTHING_TO_DO"]
                device_num = int(choice) - 1
                if 0 <= device_num < len(updates_list):
                    if not downgrade:
                        return device_num
                    break
                else:
                    raise ValueError()
            except ValueError:
                print("Invalid choice.")

        if downgrade:
            while True:
                try:
                    releases = updates_list[device_num]["Releases"]
                    for i, fw_dngd in enumerate(releases):
                        print(decorator)
                        print(
                            f"  {i+1}. Firmware downgrade version:"
                            f"\t {fw_dngd['Version']}"
                        )
                        description = fw_dngd["Description"].replace("<p>", "")
                        description = description.replace("<li>", "")
                        description = description.replace("<ul>", "")
                        description = description.replace("</ul>", "")
                        description = description.replace("</p>", "\n   ")
                        description = description.replace("</li>", "\n   ")
                        print(f"   Description:{description}")
                    print("If you want to abandon downgrade process press N.")
                    choice = input("Otherwise choose downgrade number: ")
                    if choice == "N" or choice == "n":
                        return -EXIT_CODES["NOTHING_TO_DO"]
                    downgrade_num = int(choice) - 1
                    if 0 <= downgrade_num < len(releases):
                        return device_num, downgrade_num
                    else:
                        raise ValueError()
                except ValueError:
                    print("Invalid choice.")

    def _parse_parameters(self, updates_list, choice):
        """Parses device name, url, version and SHA256 checksum of the file list.

        Keywords arguments:
        updates_list - list of updates for dom0
        choice -- number of device to be updated
        """
        self.name = updates_list[choice]["Name"]
        self.version = updates_list[choice]["Releases"][0]["Version"]
        for ver_check in updates_list[choice]["Releases"]:
            if pversion.parse(ver_check["Version"]) >= pversion.parse(self.version):
                self.version = ver_check["Version"]
                self.url = ver_check["Url"]
                self.sha = ver_check["Checksum"]

    def _install_dom0_firmware_update(self, arch_path):
        """Installs firmware update for specified device in dom0.

        Keywords arguments:
        arch_path - absolute path to firmware update archive
        """
        cmd_install = [FWUPDMGR, "install", arch_path]
        p = subprocess.Popen(cmd_install)
        p.wait()
        if p.returncode != 0:
            raise Exception("fwupd-qubes: Firmware update failed")

    def _read_dmi(self):
        """Reads BIOS information from DMI."""
        cmd_dmidecode_version = ["dmidecode", "-s", "bios-version"]
        p = subprocess.Popen(cmd_dmidecode_version, stdout=subprocess.PIPE)
        p.wait()
        self.dmi_version = p.communicate()[0].decode()
        cmd_dmidecode = ["dmidecode", "-t", "bios"]
        p = subprocess.Popen(cmd_dmidecode, stdout=subprocess.PIPE)
        p.wait()
        if p.returncode != 0:
            raise Exception("dmidecode: Reading DMI failed")
        return p.communicate()[0].decode()

    def _verify_dmi(self, arch_path, version, downgrade=False):
        """Verifies DMI tables for BIOS updates.

        Keywords arguments:
        arch_path -- absolute path of the update archive
        version -- version of the update
        downgrade -- downgrade flag
        """
        dmi_info = self._read_dmi()
        with tempfile.TemporaryDirectory() as tmpdir:
            cmd_extract = ["gcab", "-x", f"--directory={tmpdir}", "--", arch_path]
            p = subprocess.Popen(cmd_extract, stdout=subprocess.PIPE)
            p.communicate()
            if p.returncode != 0:
                raise Exception(f"gcab: Error while extracting {arch_path}.")
            path_metainfo = os.path.join(tmpdir, "firmware.metainfo.xml")
            tree = ET.parse(path_metainfo)
        root = tree.getroot()
        vendor = root.find("developer_name").text
        if vendor is None:
            raise ValueError("No vendor information in firmware metainfo.")
        if vendor not in dmi_info:
            raise ValueError("Wrong firmware provider.")
        if not downgrade and pversion.parse(version) <= pversion.parse(
            self.dmi_version
        ):
            raise ValueError(f"{version} < {self.dmi_version} Downgrade not allowed")

    def _get_dom0_devices(self):
        """Gathers information about devices connected in dom0."""
        cmd_get_dom0_devices = [FWUPDMGR, "--json", "get-devices"]
        p = subprocess.Popen(cmd_get_dom0_devices, stdout=subprocess.PIPE)
        self.dom0_devices_info = p.communicate()[0].decode()
        if p.returncode != 0:
            raise Exception("fwupd-qubes: Getting devices info failed")

    def update_firmware(self, whonix=False):
        """Updates firmware of the specified device.

        Keyword arguments:
        whonix -- Flag enforces downloading the metadata updates via Tor
        """
        self._get_dom0_updates()
        self._parse_dom0_updates_info(self.dom0_updates_info)
        updates_list = self.dom0_updates_list
        ret_input = self._user_input(updates_list)
        if ret_input == -EXIT_CODES["NOTHING_TO_DO"]:
            exit(EXIT_CODES["NOTHING_TO_DO"])
        choice = ret_input
        self._parse_parameters(updates_list, choice)
        self._download_firmware_updates(self.url, self.sha, whonix=whonix)
        if self.name == "System Firmware":
            Path(BIOS_UPDATE_FLAG).touch(mode=0o644, exist_ok=True)
            self._verify_dmi(self.arch_path, self.version)
        self._install_dom0_firmware_update(self.arch_path)

    def _parse_downgrades(self, device_list):
        """Parses information about possible downgrades.

        Keywords argument:
        device_list -- list of connected devices
        """
        downgrades = []
        if "No detected devices" in device_list:
            return downgrades
        dom0_devices_info_dict = json.loads(device_list)
        for device in dom0_devices_info_dict["Devices"]:
            if "Releases" in device:
                try:
                    version = device["Version"]
                except KeyError:
                    continue
                downgrades.append(
                    {
                        "Name": device["Name"],
                        "Version": device["Version"],
                        "Releases": [
                            {
                                "Version": downgrade["Version"],
                                "Description": downgrade["Description"],
                                "Url": downgrade["Uri"],
                                "Checksum": downgrade["Checksum"][-1],
                            }
                            for downgrade in device["Releases"]
                            if pversion.parse(downgrade["Version"])
                            < pversion.parse(version)
                        ],
                    }
                )
        return downgrades

    def _install_dom0_firmware_downgrade(self, arch_path):
        """Installs firmware downgrade for specified device.

        Keywords arguments:
        arch_path - absolute path to firmware downgrade archive
        """
        cmd_install = [FWUPDMGR, "--allow-older", "install", arch_path]
        p = subprocess.Popen(cmd_install)
        p.wait()
        if p.returncode != 0:
            raise Exception("fwupd-qubes: Firmware downgrade failed")

    def downgrade_firmware(self, whonix=False):
        """Downgrades firmware of the specified device.

        Keyword arguments:
        whonix -- Flag enforces downloading the metadata updates via Tor
        """
        self._get_dom0_devices()
        dom0_downgrades = self._parse_downgrades(self.dom0_devices_info)
        ret_input = self._user_input(dom0_downgrades, downgrade=True)
        if ret_input == -EXIT_CODES["NOTHING_TO_DO"]:
            exit(EXIT_CODES["NOTHING_TO_DO"])
        device_choice, downgrade_choice = ret_input
        downgrade = dom0_downgrades[device_choice]
        releases = downgrade["Releases"]
        downgrade_url = releases[downgrade_choice]["Url"]
        downgrade_sha = releases[downgrade_choice]["Checksum"]
        self._download_firmware_updates(downgrade_url, downgrade_sha, whonix=whonix)
        if downgrade["Name"] == "System Firmware":
            Path(BIOS_UPDATE_FLAG).touch(mode=0o644, exist_ok=True)
            self._verify_dmi(
                self.arch_path,
                downgrade["Version"],
                downgrade=True,
            )
        self._install_dom0_firmware_downgrade(self.arch_path)

    def _output_crawler(self, updev_dict, level, help_f=False):
        """Prints device and updates information as a tree.

        Keywords arguments:
        updev_dict -- update/device information dictionary
        level -- level of the tree
        """

        def _tabs(key_word):
            return key_word + "\t" * (4 - int(len(key_word) / 8))

        decorator = "==================================="
        print(2 * decorator)
        for updev_key in updev_dict:
            style = "\t" * level
            output = style + _tabs(updev_key + ":")
            if len(updev_key) > 12:
                continue
            if updev_key == "Icons":
                continue
            if updev_key == "Releases":
                continue
            if updev_key == "Name":
                print(style + updev_dict["Name"])
                print(2 * decorator)
                continue
            if isinstance(updev_dict[updev_key], str):
                print(output + updev_dict[updev_key])
            elif isinstance(updev_dict[updev_key], int):
                print(output + str(updev_dict[updev_key]))
            elif isinstance(updev_dict[updev_key][0], str):
                for i, data in enumerate(updev_dict[updev_key]):
                    if i == 0:
                        print(output + "\u00B7" + data)
                        continue
                    print(style + _tabs(" ") + "\u00B7" + data)
            elif isinstance(updev_dict[updev_key][0], dict):
                if level == 0 and help_f is True:
                    print(output)
                else:
                    if level == 0:
                        print(f"Dom0 {output}")

                for nested_dict in updev_dict[updev_key]:
                    self._output_crawler(nested_dict, level + 1)

    def _updates_crawler(self, updates_list, prefix=0):
        """Prints updates information for dom0

        Keywords arguments:
        updates_list -- list of devices updates
        prefix -- device number prefix
        """
        available_updates = False
        decorator = "======================================================"
        print(decorator)
        print("Dom0 updates:")
        print(decorator)
        if len(updates_list) == 0:
            print("No updates available.")
            return EXIT_CODES["NOTHING_TO_DO"]
        else:
            for i, device in enumerate(updates_list):
                if len(device["Releases"]) == 0:
                    continue
                if not available_updates:
                    print("Available updates:")
                    print(decorator)
                print("^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^")
                print(f"{i+1+prefix}. Device: {device['Name']}")
                print(f"   Current firmware version:\t {device['Version']}")
                for update in device["Releases"]:
                    print(decorator)
                    print("   Firmware update " f"version:\t {update['Version']}")
                    print(f"   URL:\t {update['Url']}")
                    print(f"   SHA256 checksum:\t {update['Checksum']}")
                    description = update["Description"].replace("<p>", "")
                    description = description.replace("<li>", "")
                    description = description.replace("<ul>", "")
                    description = description.replace("</ul>", "")
                    description = description.replace("</p>", "\n\t")
                    description = description.replace("</li>", "\n\t")
                    print(f"   Description: {description}")
                print(decorator)
                available_updates = True
            if not available_updates:
                print("No updates available.")
                return EXIT_CODES["NOTHING_TO_DO"]

    def get_devices_qubes(self):
        """Gathers and prints devices information."""
        self._get_dom0_devices()
        dom0_devices_info_dict = json.loads(self.dom0_devices_info)
        self._output_crawler(dom0_devices_info_dict, 0)

    def get_updates_qubes(self):
        """Gathers and prints updates information."""
        self._get_dom0_updates()
        self._parse_dom0_updates_info(self.dom0_updates_info)
        self._updates_crawler(self.dom0_updates_list)

    def help(self):
        """Prints help information"""
        self._output_crawler(HELP, 0, help_f=True)

    def trusted_cleanup(self):
        """Deletes trusted directory."""
        trusted_path = os.path.join(FWUPD_DOM0_UPDATES_DIR, "trusted.cab")
        if os.path.exists(trusted_path):
            os.remove(trusted_path)
            shutil.rmtree(trusted_path.replace(".cab", ""))

    def refresh_metadata_after_bios_update(self):
        """Refreshes metadata after bios update"""
        if os.path.exists(BIOS_UPDATE_FLAG):
            print("BIOS was updated. Refreshing metadata...")
            if "--whonix" in sys.argv:
                self.refresh_metadata_all(whonix=True)
            else:
                self.refresh_metadata_all()
            os.remove(BIOS_UPDATE_FLAG)

    def heads_update(self, device=None, whonix=False, metadata_url=None):
        """
        Updates heads firmware

        Keyword arguments:
        device -- Model of the updated device
        whonix -- Flag enforces downloading the metadata updates via Tor
        metadata_url -- Use custom metadata from the url
        """
        if not metadata_url:
            metadata_url = METADATA_URL
        metadata_name = os.path.basename(metadata_url)
        self.metadata_file = os.path.join(FWUPD_DOM0_METADATA_DIR, metadata_name)
        self._get_hwids()
        if device is None:
            device = self._get_hwid_device()
        self._download_metadata(whonix=whonix, metadata_url=metadata_url)
        self._parse_metadata(self.metadata_file)
        if self._gather_firmware_version() == EXIT_CODES["NOTHING_TO_DO"]:
            return EXIT_CODES["NOTHING_TO_DO"]
        if self._parse_heads_updates(device) == EXIT_CODES["NOTHING_TO_DO"]:
            return EXIT_CODES["NOTHING_TO_DO"]
        self._download_firmware_updates(self.heads_update_url, self.heads_update_sha)
        return_code = self._copy_heads_firmware(self.arch_path)
        if return_code == EXIT_CODES["NOTHING_TO_DO"]:
            exit(EXIT_CODES["NOTHING_TO_DO"])
        elif return_code == EXIT_CODES["SUCCESS"]:
            print()
            while True:
                try:
                    print("An update requires a reboot to complete.")
                    choice = input("Do you want to restart now? (Y|N)\n")
                    if choice == "N" or choice == "n":
                        return EXIT_CODES["SUCCESS"]
                    elif choice == "Y" or choice == "y":
                        print("Rebooting...")
                        os.system("reboot")
                    else:
                        raise ValueError()
                except ValueError:
                    print("Invalid choice.")
        else:
            raise Exception("Copying heads update failed!!")

    def validate_dom0_dirs(self):
        """Validates and creates directories"""
        if not os.path.exists(FWUPD_DOM0_DIR):
            create_dirs(FWUPD_DOM0_DIR)
        if os.path.exists(FWUPD_DOM0_METADATA_DIR):
            shutil.rmtree(FWUPD_DOM0_METADATA_DIR)
            create_dirs(FWUPD_DOM0_METADATA_DIR)
        else:
            create_dirs(FWUPD_DOM0_METADATA_DIR)
        if not os.path.exists(FWUPD_DOM0_UPDATES_DIR):
            create_dirs(FWUPD_DOM0_UPDATES_DIR)


def main():
    if os.geteuid() != 0:
        print("You need to have root privileges to run this script.\n")
        exit(EXIT_CODES["ERROR"])

    q = QubesFwupdmgr()

    if len(sys.argv) < 2:
        q.help()
        exit(1)

    metadata_url = None
    device_override = None
    whonix = False

    for arg in sys.argv:
        if "--url=" in arg:
            metadata_url = arg.replace("--url=", "")
            if FWUPD_DOWNLOAD_PREFIX not in metadata_url:
                print(
                    "Metadata must be stored in the Linux"
                    " Vendor Firmware Service (https://fwupd.org/)"
                )
                print("Exiting...")
                exit(1)
        if "--device=" in arg:
            device_override = arg.replace("--device=", "")
        if "--whonix" == arg:
            whonix = True

    q.validate_dom0_dirs()
    q.trusted_cleanup()
    q.refresh_metadata_after_bios_update()

    if not os.path.exists(FWUPD_DOM0_DIR):
        if metadata_url:
            q.refresh_metadata(whonix=whonix, metadata_url=metadata_url)
        else:
            q.refresh_metadata_all(whonix=whonix)

    if sys.argv[1] == "get-updates":
        q.get_updates_qubes()
    elif sys.argv[1] == "get-devices":
        q.get_devices_qubes()
    elif sys.argv[1] == "update":
        q.update_firmware(whonix=whonix)
    elif sys.argv[1] == "downgrade":
        q.downgrade_firmware(whonix=whonix)
    elif sys.argv[1] == "clean":
        q.clean_cache()
    elif sys.argv[1] == "refresh":
        if metadata_url:
            q.refresh_metadata(whonix=whonix, metadata_url=metadata_url)
        else:
            q.refresh_metadata_all(whonix=whonix)
    elif sys.argv[1] == "update-heads":
        q.heads_update(device=device_override, metadata_url=metadata_url, whonix=whonix)
    else:
        q.help()
        exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/python3
#
# Copyright (C) 2023 Richard Hughes <richard@hughsie.com>
#
# SPDX-License-Identifier: LGPL-2.1+
#
# pylint: disable=invalid-name,missing-docstring,consider-using-f-string

import os
import datetime
import argparse
import glob
import sys
from jinja2 import Environment, FileSystemLoader, select_autoescape

subst = {}


def _fix_case(value: str) -> str:

    return value[0].upper() + value[1:].lower()


def _subst_add_string(key: str, value: str) -> None:

    # sanity check
    if not value.isascii():
        raise NotImplementedError(f"{key} can only be ASCII, got {value}")
    if len(value) < 3:
        raise NotImplementedError(f"{key} has to be at least length 3, got {value}")

    subst[key] = value
    subst[key.lower()] = value.lower()
    subst[key.upper()] = value.upper()


def _subst_replace(data: str) -> str:

    for key, value in subst.items():
        data = data.replace(key, value)
    return data


if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--vendor",
        type=str,
        help="Vendor name",
        required=True,
    )
    parser.add_argument(
        "--example",
        type=str,
        help="Plugin basename",
        required=True,
    )
    parser.add_argument("--parent", type=str, default="Usb", help="Device parent GType")
    parser.add_argument(
        "--year", type=int, default=datetime.date.today().year, help="Copyright year"
    )
    parser.add_argument("--author", type=str, help="Copyright author", required=True)
    parser.add_argument("--email", type=str, help="Copyright email", required=True)
    args = parser.parse_args()

    try:
        _subst_add_string("Vendor", _fix_case(args.vendor))
        _subst_add_string("Example", _fix_case(args.example))
        _subst_add_string("Parent", args.parent)
        _subst_add_string("Year", str(args.year))
        _subst_add_string("Author", args.author)
        _subst_add_string("Email", args.email)
    except NotImplementedError as e:
        print(e)
        sys.exit(1)

    template_src = "vendor-example"
    os.makedirs(os.path.join("plugins", _subst_replace(template_src)), exist_ok=True)

    env = Environment(
        loader=FileSystemLoader("."),
        autoescape=select_autoescape(),
        keep_trailing_newline=True,
    )
    for fn in glob.iglob(f"./plugins/{template_src}/*"):
        template = env.get_template(fn)
        with open(_subst_replace(fn.replace(".in", "")), "wb") as f_dst:
            f_dst.write(template.render(subst).encode())



#!/usr/bin/python3
# pylint: disable=invalid-name,missing-docstring,consider-using-f-string
#
# Copyright (C) 2023 Richard Hughes <richard@hughsie.com>
#
# SPDX-License-Identifier: LGPL-2.1+

import os
import sys
import argparse
from typing import List, Dict
from jinja2 import Environment, FileSystemLoader, select_autoescape


def _replace_bookend(line: str, search: str, replace_l: str, replace_r: str) -> str:

    try:
        while line.find(search) != -1:
            it = iter(line.split(search, maxsplit=2))
            line_tmp: str = ""
            for token_before in it:
                line_tmp += token_before
                line_tmp += replace_l
                line_tmp += next(it)  # token_mid
                line_tmp += replace_r
                line_tmp += next(it)  # token_after
            line = line_tmp
    except StopIteration:
        pass
    return line


def _strip_md(data: str) -> str:
    content = ""
    for line in data.split("\n"):
        # skip the man page header
        if line.startswith("%"):
            continue
        if line.startswith("|"):
            line = line[2:]
        # create links to other "man" pages
        if line.startswith("<") and (line.endswith("(1)>") or line.endswith("(5)>")):
            line = line.strip("<>")
            name = line.split("(")[0]
            line = f"[`{line}`](./{name}.html)"
        content += f"{line}\n"
    return content


def _convert_md_to_man(data: str) -> str:

    sections = data.split("\n\n")
    troff_lines: List[str] = []

    # ignore the docgen header
    if sections[0].startswith("---"):
        sections = sections[1:]

    # header
    split = sections[0].split(" ", maxsplit=4)
    if split[0] != "%" or split[3] != "|":
        print(
            "no man header detected, expected something like "
            "'% fwupdagent(1) 1.2.5 | fwupdagent man page' and got {}".format(
                sections[0]
            )
        )
        sys.exit(1)
    man_cmd = split[1][:-3]
    man_sect = int(split[1][-2:-1])
    troff_lines.append(f'.TH "{man_cmd}" "{man_sect}" "" {split[2]} "{split[4]}"')
    troff_lines.append(".hy")  # hyphenate

    # content
    for section in sections[1:]:
        lines = section.split("\n")
        sectkind: str = ".PP"  # begin a new paragraph
        sectalign: int = 4

        # convert markdown headers to section headers
        if lines[-1].startswith("##"):
            lines = [lines[-1].strip("#").strip()]
            sectkind = ".SH"

        # join long lines
        line = ""
        indent = False
        for line_tmp in lines:
            if not line_tmp:
                continue
            if line_tmp.startswith("```"):
                indent = not indent
                line_tmp = "```"  # strip the language
            if line_tmp.startswith("| "):
                line_tmp = line_tmp[2:]
            if indent:
                line += ".nf\n"
                line += line_tmp + "\n"
                line += ".fi\n"
                continue
            elif line_tmp.startswith("* "):
                line += ".IP \\[bu] 2\n"
                line += line_tmp[2:]
                continue
            line_tmp = line_tmp.replace("\\-", "-")
            line += line_tmp
            if line_tmp.endswith("."):
                line += " "
            line += " "

        # make bold, and add smart quotes
        line = _replace_bookend(line, "**", "\\f[B]", "\\f[R]")
        line = _replace_bookend(line, "`", "\\f[B]", "\\f[R]")
        line = _replace_bookend(line, '"', "“", "”")

        # add troff
        if sectalign != 4:
            troff_lines.append(f".RS {sectalign}")
        troff_lines.append(sectkind)
        troff_lines.append(line)
        if sectalign != 4:
            troff_lines.append(".RE")

    # success
    return "\n".join(troff_lines)


def _add_defines(defines: Dict[str, str], fn: str) -> None:
    with open(fn, "rb") as f:
        for line in f.read().decode().split("\n"):
            if (
                line.find("set_config_default") == -1
                and line.find("config_set_default") == -1
            ):
                continue
            try:
                _, wrapped_key, wrapped_value = line.split(",", maxsplit=2)
                wrapped_value = wrapped_value.rsplit(")", maxsplit=1)[0]
            except ValueError:
                print(f"failed to define value for {line} in {fn}")
                continue
            try:
                _, key, _ = wrapped_key.split('"', maxsplit=2)
            except ValueError:
                continue
            try:
                _, value, _ = wrapped_value.split('"', maxsplit=2)
            except ValueError:
                value = wrapped_value.strip()
            source_prefix: str = (
                os.path.basename(os.path.dirname(fn)).replace("-", "_") + "_"
            )
            if source_prefix == "src_":
                source_prefix = ""
            defines[f"{source_prefix}{key}"] = {"NULL": ""}.get(value, value)


if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument("-o", "--output")
    parser.add_argument(
        "-r", "--replace", action="append", nargs=2, metavar=("symbol", "version")
    )
    parser.add_argument("-d", "--defines", action="append")
    parser.add_argument("--md", action="store_true")
    args, argv = parser.parse_known_args()
    if len(argv) != 1:
        print(f"usage: {sys.argv[0]} MARKDOWN [-o TROFF]\n")
        sys.exit(1)

    # load in #defines to populate the defaults
    subst: Dict[str, str] = {}
    if args.defines:
        for fn_define in args.defines:
            try:
                _add_defines(subst, fn_define)
            except FileNotFoundError:
                print(f"{fn_define} not found")
                sys.exit(1)

    # static defines
    if args.replace:
        for key, value in args.replace:
            subst[key] = value

    # use Jinja2 to process as a template
    env = Environment(
        loader=FileSystemLoader(os.path.dirname(argv[0])),
        autoescape=select_autoescape(),
        keep_trailing_newline=True,
    )
    template = env.get_template(os.path.basename(argv[0]))
    rendered = template.render(subst)

    # Stripped markdown mode is used for HTML docs
    if args.md:
        out = _strip_md(rendered)
    else:
        out = _convert_md_to_man(rendered)

    # success
    if args.output:
        with open(args.output, "wb") as f_out:
            f_out.write(out.encode())
    else:
        print(out)



#!/usr/bin/python3
#
# The Qubes OS Project, http://www.qubes-os.org
#
# Copyright (C) 2021  Norbert Kamiński  <norbert.kaminski@3mdeb.com>
#
# SPDX-License-Identifier: LGPL-2.1+
#

import subprocess
import tempfile
import os
import shutil
import xml.etree.ElementTree as ET
from packaging.version import Version
from qubes_fwupd_common import EXIT_CODES, create_dirs, LooseVersion

FWUPDTOOL = "/bin/fwupdtool"

BOOT = "/boot"
HEADS_UPDATES_DIR = os.path.join(BOOT, "updates")


class FwupdHeads:
    def _get_hwids(self):
        cmd_hwids = [FWUPDTOOL, "hwids"]
        p = subprocess.Popen(cmd_hwids, stdout=subprocess.PIPE)
        self.dom0_hwids_info = p.communicate()[0].decode()
        if p.returncode != 0:
            raise Exception("fwupd-qubes: Getting hwids info failed")

    def _gather_firmware_version(self):
        """
        Checks if Qubes works under heads
        """
        if "Heads" in self.dom0_hwids_info:
            self.heads_version = None
            hwids = self.dom0_hwids_info.split("\n")
            for line in hwids:
                if "Heads" in line:
                    self.heads_version = line.split("Heads-v")[1]
        else:
            print("Device is not running under the heads firmware!!")
            print("Exiting...")
            return EXIT_CODES["NOTHING_TO_DO"]

    def _get_hwid_device(self):
        """
        Device model for Heads update, currently supports ThinkPad only.
        """
        for line in self.dom0_hwids_info.splitlines():
            if line.startswith("Family: ThinkPad"):
                return line.split(":", 1)[1].split(" ", 1)[1].lower()
        return None

    def _parse_metadata(self, metadata_file):
        """
        Parse metadata info.
        """
        metadata_ext = os.path.splitext(metadata_file)[-1]
        if metadata_ext == ".xz":
            cmd_metadata = ["xzcat", metadata_file]
        elif metadata_ext == ".gz":
            cmd_metadata = ["zcat", metadata_file]
        else:
            raise NotImplementedError(
                "Unsupported metadata compression " + metadata_ext
            )
        p = subprocess.Popen(cmd_metadata, stdout=subprocess.PIPE)
        self.metadata_info = p.communicate()[0].decode()
        if p.returncode != 0:
            raise Exception("fwupd-qubes: Parsing metadata failed")

    def _parse_heads_updates(self, device):
        """
        Parses heads updates info.

        Keyword arguments:
        device -- Model of the updated device
        """
        self.heads_update_url = None
        self.heads_update_sha = None
        self.heads_update_version = None
        heads_metadata_info = None
        root = ET.fromstring(self.metadata_info)
        for component in root.findall("component"):
            if f"heads.{device}" in component.find("id").text:
                heads_metadata_info = component
        if not heads_metadata_info:
            print("No metadata info for chosen board")
            return EXIT_CODES["NOTHING_TO_DO"]
        for release in heads_metadata_info.find("releases").findall("release"):
            release_ver = release.get("version")
            if self.heads_version == "heads" or LooseVersion(
                release_ver
            ) > LooseVersion(self.heads_version):
                if not self.heads_update_version or LooseVersion(
                    release_ver
                ) > LooseVersion(self.heads_update_version):
                    self.heads_update_url = release.find("location").text
                    for sha in release.findall("checksum"):
                        if (
                            ".cab" in sha.attrib["filename"]
                            and sha.attrib["type"] == "sha256"
                        ):
                            self.heads_update_sha = sha.text
                    self.heads_update_version = release_ver
        if self.heads_update_url:
            return EXIT_CODES["SUCCESS"]
        else:
            print("Heads firmware is up to date.")
            return EXIT_CODES["NOTHING_TO_DO"]

    def _copy_heads_firmware(self, arch_path):
        """
        Copies heads update to the boot path
        """
        heads_boot_path = os.path.join(HEADS_UPDATES_DIR, self.heads_update_version)

        heads_update_path = os.path.join(heads_boot_path, "firmware.rom")
        create_dirs(HEADS_UPDATES_DIR)
        if os.path.exists(heads_update_path):
            print(f"Heads Update == {self.heads_update_version} " "already downloaded.")
            return EXIT_CODES["NOTHING_TO_DO"]
        else:
            os.mkdir(heads_boot_path)
            with tempfile.TemporaryDirectory() as tmpdir:
                cmd_extract = ["gcab", "-x", f"--directory={tmpdir}", "--", arch_path]
                p = subprocess.Popen(cmd_extract, stdout=subprocess.PIPE)
                p.communicate()
                if p.returncode != 0:
                    raise Exception(f"gcab: Error while extracting {arch_path}.")
                update_path = os.path.join(tmpdir, "firmware.rom")
                shutil.copyfile(update_path, heads_update_path)
            print(
                f"Heads Update == {self.heads_update_version} "
                f"available at {heads_boot_path}"
            )
            return EXIT_CODES["SUCCESS"]



#!/usr/bin/python3
#
# Copyright (C) 2017 Dell, Inc.
#
# SPDX-License-Identifier: LGPL-2.1+
#
from base64 import b64decode
import io
import os
import subprocess
import sys
import shutil
import tempfile
import zipfile

TAG = b"#\x00"


def parse_args():
    import argparse

    parser = argparse.ArgumentParser(description="Self extracting firmware updater")
    parser.add_argument("--directory", help="Directory to extract to")
    parser.add_argument(
        "--cleanup",
        action="store_true",
        help="Remove tools when done with installation",
    )
    parser.add_argument(
        "--verbose", action="store_true", help="Run the tool in verbose mode"
    )
    parser.add_argument(
        "--allow-reinstall",
        action="store_true",
        help="Allow re-installing existing firmware versions",
    )
    parser.add_argument(
        "--allow-older", action="store_true", help="Allow downgrading firmware versions"
    )
    parser.add_argument(
        "command", choices=["install", "extract"], help="Command to run"
    )
    args = parser.parse_args()
    return args


def error(msg):
    print(msg)
    sys.exit(1)


def bytes_slicer(length, source):
    start = 0
    stop = length
    while start < len(source):
        yield source[start:stop]
        start = stop
        stop += length


def get_zip():
    script = os.path.realpath(__file__)
    bytes_out = io.BytesIO()
    with open(script, "rb") as source:
        for line in source:
            if not line.startswith(TAG):
                continue
            bytes_out.write(b64decode(line[len(TAG) : -1]))
    return bytes_out


def unzip(destination):
    zipf = get_zip()
    source = zipfile.ZipFile(zipf, "r")
    for item in source.namelist():
        # extract handles the sanitization
        source.extract(item, destination)


def copy_cabs(source, target):
    if not os.path.exists(target):
        os.makedirs(target)
    cabs = []
    for root, dirs, files in os.walk(source):
        for f in files:
            if f.endswith(".cab"):
                origf = os.path.join(root, f)
                shutil.copy(origf, target)
                cabs.append(os.path.join(target, f))
    return cabs


def install_snap(directory, verbose, allow_reinstall, allow_older, uninstall):
    app = "fwupd"
    common = f"/root/snap/{app}/common"

    # check if snap is installed
    with open(os.devnull, "w") as devnull:
        subprocess.run(["snap"], check=True, stdout=devnull, stderr=devnull)

    # check existing installed
    cmd = ["snap", "list", app]
    with open(os.devnull, "w") as devnull:
        if verbose:
            print(cmd)
        ret = subprocess.run(cmd, stdout=devnull, stderr=devnull)
        if ret.returncode == 0:
            cmd = ["snap", "remove", app]
            if verbose:
                print(cmd)
            subprocess.run(cmd, check=True)

    # install the snap
    cmd = ["snap", "ack", os.path.join(directory, "fwupd.assert")]
    if verbose:
        print(cmd)
    subprocess.run(cmd, check=True)
    cmd = ["snap", "install", "--classic", os.path.join(directory, "fwupd.snap")]
    if verbose:
        print(cmd)
    subprocess.run(cmd, check=True)

    # copy the CAB files
    cabs = copy_cabs(directory, common)

    # run the snap
    for cab in cabs:
        cmd = [f"{app}.fwupdmgr", "install", cab]
        if allow_reinstall:
            cmd += ["--allow-reinstall"]
        if allow_older:
            cmd += ["--allow-older"]
        if verbose:
            cmd += ["--verbose"]
            print(cmd)
        subprocess.run(cmd)

    # remove copied cabs
    for f in cabs:
        os.remove(f)

    # cleanup
    if uninstall:
        cmd = ["snap", "remove", app]
        if verbose:
            print(cmd)
        subprocess.run(cmd)


def install_flatpak(directory, verbose, allow_reinstall, allow_older, uninstall):
    app = "org.freedesktop.fwupd"
    common = f"{os.getenv('HOME')}/.var/app/{app}"

    with open(os.devnull, "w") as devnull:
        if not verbose:
            output = devnull
        else:
            output = None
        # look for dependencies
        dep = "org.gnome.Platform/x86_64/3.30"
        repo = "flathub"
        repo_url = "https://flathub.org/repo/flathub.flatpakrepo"
        cmd = ["flatpak", "info", dep]
        if verbose:
            print(cmd)
        ret = subprocess.run(cmd, stdout=output, stderr=output)
        # not installed
        if ret.returncode != 0:
            # look for remotes
            cmd = ["flatpak", "remote-info", repo, dep]
            if verbose:
                print(cmd)
            ret = subprocess.run(cmd, stdout=output, stderr=output)
            # not enabled, enable it
            if ret.returncode != 0:
                cmd = ["flatpak", "remote-add", repo, repo_url]
                if verbose:
                    print(cmd)
                ret = subprocess.run(cmd, stderr=output)
            # install dep
            cmd = ["flatpak", "install", repo, dep]
            if verbose:
                print(cmd)
            ret = subprocess.run(cmd)

        # check existing installed
        cmd = ["flatpak", "info", app]
        if verbose:
            print(cmd)
        ret = subprocess.run(cmd, stdout=output, stderr=output)
        if ret.returncode == 0:
            cmd = ["flatpak", "remove", app]
            if verbose:
                print(cmd)
            subprocess.run(cmd, check=True)

    # install the flatpak
    cmd = ["flatpak", "install", os.path.join(directory, "fwupd.flatpak")]
    if verbose:
        print(cmd)
    subprocess.run(cmd, check=True)

    # copy the CAB files
    cabs = copy_cabs(directory, common)

    # run command
    for cab in cabs:
        cmd = ["flatpak", "run", app, "install", cab]
        if allow_reinstall:
            cmd += ["--allow-reinstall"]
        if allow_older:
            cmd += ["--allow-older"]
        if verbose:
            cmd += ["--verbose"]
            print(cmd)
        subprocess.run(cmd)

    # remove copied cabs
    for f in cabs:
        os.remove(f)

    # cleanup
    if uninstall:
        cmd = ["flatpak", "remove", app]
        if verbose:
            print(cmd)
        subprocess.run(cmd)


# Check which package to use
# - return False to use packaged version
# - return True for snap/flatpak
def use_included_version(minimum_version):
    try:
        import apt
    except ModuleNotFoundError:
        return True
    cache = apt.Cache()
    pkg = cache.get("fwupd")
    version = pkg.installed
    if not version:
        return True
    if minimum_version:
        if minimum_version > version:
            print(
                "fwupd %s is already installed but this package requires %s"
                % (version.version, minimum_version)
            )
        else:
            print(
                "Using existing fwupd version %s already installed on system."
                % version.version
            )
            return False
    else:
        print(f"fwupd {version.version} is installed and must be removed")
    return remove_packaged_version(pkg, cache)


def remove_packaged_version(pkg, cache):
    res = False
    while True:
        res = input("Remove now (Y/N)? ")
        if res.lower() == "n":
            res = False
            break
        if res.lower() == "y":
            res = True
            break
    if res:
        pkg.mark_delete()
        res = cache.commit()
    if not res:
        raise Exception("Need to remove packaged version")
    return True


def install_builtin(directory, verbose, allow_reinstall, allow_older):
    cabs = []
    for root, dirs, files in os.walk(directory):
        for f in files:
            if f.endswith(".cab"):
                cabs.append(os.path.join(root, f))
    # run command
    for cab in cabs:
        cmd = ["fwupdmgr", "install", cab]
        if allow_reinstall:
            cmd += ["--allow-reinstall"]
        if allow_older:
            cmd += ["--allow-older"]
        if verbose:
            cmd += ["--verbose"]
            print(cmd)
        subprocess.run(cmd)


def run_installation(directory, verbose, allow_reinstall, allow_older, uninstall):
    try_snap = False
    try_flatpak = False

    # determine if a minimum version was specified
    minimum_path = os.path.join(directory, "minimum")
    minimum = None
    if os.path.exists(minimum_path):
        with open(minimum_path, "r") as rfd:
            minimum = rfd.read()

    if not use_included_version(minimum):
        install_builtin(directory, verbose, allow_reinstall, allow_older)
        return

    # determine what self extracting binary has
    if os.path.exists(os.path.join(directory, "fwupd.snap")) and os.path.exists(
        os.path.join(directory, "fwupd.assert")
    ):
        try_snap = True
    if os.path.exists(os.path.join(directory, "fwupd.flatpak")):
        try_flatpak = True

    if try_snap:
        try:
            install_snap(directory, verbose, allow_reinstall, allow_older, uninstall)
            return True
        except Exception:
            if verbose:
                print("Snap installation failed")
            if not try_flatpak:
                error("Snap installation failed")
    if try_flatpak:
        install_flatpak(directory, verbose, allow_reinstall, allow_older, uninstall)


if __name__ == "__main__":
    args = parse_args()
    if "extract" in args.command:
        if args.allow_reinstall:
            error(
                "allow-reinstall argument doesn't make sense with command %s"
                % args.command
            )
        if args.allow_older:
            error(
                f"allow-older argument doesn't make sense with command {args.command}"
            )
        if args.cleanup:
            error(f"Cleanup argument doesn't make sense with command {args.command}")
        if args.directory is None:
            error("No directory specified")
        if not os.path.exists(args.directory):
            print(f"Creating {args.directory}")
            os.makedirs(args.directory)
        unzip(args.directory)
    else:
        if args.directory:
            error(
                "Directory argument %s doesn't make sense with command %s"
                % (args.directory, args.command)
            )
        if os.getuid() != 0:
            error("This tool must be run as root")
        with tempfile.TemporaryDirectory(prefix="fwupd") as target:
            unzip(target)
            run_installation(
                target,
                args.verbose,
                args.allow_reinstall,
                args.allow_older,
                args.cleanup,
            )



#!/usr/bin/python3
# pylint: disable=invalid-name,missing-docstring
#
# Copyright (C) 2023 Richard Hughes <richard@hughsie.com>
#
# SPDX-License-Identifier: LGPL-2.1+

import argparse
import gzip
from typing import List

if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument("output", action="store", type=str, help="output")
    parser.add_argument("input", nargs="*", help="input")
    args = parser.parse_args()

    lines: List[str] = []
    for fn in args.input:
        with open(fn, "rb") as f:
            for line in f.read().decode().split("\n"):
                if not line:
                    continue
                if line.startswith("#"):
                    continue
                lines.append(line)
    with gzip.GzipFile(args.output, "wb", mtime=0) as f:
        f.write("\n".join(lines).encode())
